import wx

from atrip.machines.atari8bit.jumpman.parser import is_valid_level_segment

from sawx.ui.dialogs import ChooseOnePlusCustomDialog
from sawx.utils.textutil import text_to_int

from ..action import ViewerAction, ViewerActionMixin, ViewerListAction, ViewerRadioAction
from .segment import segment_select
from ..commands import jumpman as jc
from ..jumpman import mouse_modes as jm

import logging
log = logging.getLogger(__name__)


class jumpman_clear_trigger(ViewerAction):
    """Remove any trigger function from the selected coin(s).
    
    """
    name = "Clear Trigger Function"
    enabled_name = 'can_copy'
    command = jc.ClearTriggerCommand

    def calc_enabled(self, action_key):
        obj = self.popup_data.get('jumpman_obj', None)
        if obj is not None:
            clearable = any(o.trigger_function is not None for o in obj)
        else:
            clearable = False
        return clearable

    def get_objects(self):
        obj = self.popup_data.get('jumpman_obj', None)
        if obj is not None:
            coins = [o for o in obj if o.single]
            return coins
        return self.viewer.control.mouse_mode.objects

    def get_addr(self, event, objects):
        return None

    def permute_object(self, obj, addr):
        obj.trigger_function = addr

    def perform(self, event):
        objects = self.get_objects()
        try:
            addr = self.get_addr(event, objects)
            for o in objects:
                self.permute_object(o, addr)
            self.viewer.save_changes(self.command)
            self.viewer.control.mouse_mode.resync_objects()
        except ValueError:
            pass


def trigger_dialog(segment_viewer, obj):
    model = segment_viewer.segment.jumpman_playfield_model
    possible_labels = model.get_triggers()
    label = model.get_trigger_label(obj.trigger_function)
    if label is None and obj.trigger_function:
        custom_value = "%04x" % obj.trigger_function
    else:
        custom_value = ""
    dlg = ChooseOnePlusCustomDialog(segment_viewer.control, list(possible_labels.keys()), label, custom_value, "Choose Trigger Function", "Select one trigger function or enter custom address", "Trigger Addr (hex)")
    if dlg.ShowModal() == wx.ID_OK:
        label, addr = dlg.get_selected()
        if label is not None:
            addr = possible_labels[label]
        else:
            try:
                addr = text_to_int(addr, "hex")
            except ValueError:
                segment_viewer.frame.error("Invalid address %s" % addr)
                addr = None
    else:
        addr = None
    dlg.Destroy()
    return addr


class jumpman_set_trigger(jumpman_clear_trigger):
    """Set a trigger function for the selected coin(s).

    If you have used the custom code option, have compiled your code using the
    built-in assembler, *and* your code has labels that start with ``trigger``,
    these will show up in the list that appears when you invoke this action.

    Otherwise, you can specify the hex address of a subroutine.
    """
    name = "Set Trigger Function..."
    command = jc.SetTriggerCommand

    def calc_enabled(self, action_key):
        obj = self.popup_data.get('jumpman_obj', None)
        # As long as there is any coin in the selection list, make it possible
        # to set a trigger on the coins in the selection
        if obj is not None:
            settable = any(o.single for o in obj)
        else:
            settable = False
        return settable

    def get_addr(self, event, objects):
        addr = trigger_dialog(self.viewer, objects[0])
        if addr is not None:
            return addr
        raise ValueError("Cancelled!")


class select_all(ViewerAction):
    """Select all drawing elements in the main level

    """
    name = 'Select All'
    accelerator = 'Ctrl+A'
    tooltip = 'Select the entire document'
    enabled_name = 'can_select_objects'

    def perform(self, event):
        self.viewer.select_all()


class select_none(ViewerAction):
    """Clear all selections

    """
    name = 'Select None'
    accelerator = 'Shift+Ctrl+A'
    tooltip = 'Clear selection'
    enabled_name = 'can_select_objects'

    def perform(self, event):
        self.viewer.select_none()


class select_invert(ViewerAction):
    """Invert the selection; that is: select everything that is currently
    unselected and unselect those that were selected.

    """
    name = 'Invert Selection'
    tooltip = 'Invert selection'
    enabled_name = 'can_select_objects'

    def perform(self, event):
        self.viewer.select_invert()


class flip_vertical(ViewerAction):
    """Flips the selected items top to bottom.

    This calculates the bounding box of just the selected items and uses that
    to find the centerline about which to flip.
    """
    name = "Flip Selection Vertically"
    enabled_name = 'can_copy'
    picked = None
    command = jc.FlipVerticalCommand

    def permute_object(self, obj, bounds):
        obj.flip_vertical(bounds)

    def perform(self, event):
        e = self.editor
        objects = e.bitmap.mouse_mode.objects
        bounds = DrawObjectBounds.get_bounds(objects)
        for o in e.bitmap.mouse_mode.objects:
            self.permute_object(o, bounds)
        e.bitmap.save_changes(self.command)
        e.bitmap.mouse_mode.resync_objects()


class flip_horizontal(flip_vertical):
    """Flips the selected items left to right.

    This calculates the bounding box of just the selected items and uses that
    to find the centerline about which to flip.
    """
    name = "Flip Selection Horizontally"
    command = jc.FlipHorizontalCommand

    def permute_object(self, obj, bounds):
        obj.flip_horizontal(bounds)


class jumpman_add_assembly_source(ViewerAction):
    """Add an assembly source file to this level (and compile it)

    This is used to provide custom actions or even game loops, beyond what is
    already built-in with trigger painting. There are special labels that are
    recognized by the assembler and used in the appropriate places:

        * vbi1
        * vbi2
        * vbi3
        * vbi4
        * dead_begin
        * dead_at_bottom
        * dead_falling
        * gameloop
        * out_of_lives
        * level_complete
        * collect_callback

    See our `reverse engineering notes
    <http://playermissile.com/jumpman/notes.html#h.s0ullubzr0vv>`_ for more
    details.
    """
    name = 'Custom Code...'

    def perform(self, event):
        linked_base = self.viewer.linked_base
        path = linked_base.frame.prompt_local_file_dialog("Assembly Source File")
        if path is not None:
            linked_base.segment.set_assembly_source(path)


class jumpman_compile_assembly_source(ViewerAction):
    """Recompile the assembly source code.

    This is a manual action, currently the program doesn't know when the file
    has changed. Making this process more automatic is a planned future
    enhancement.
    """
    name = 'Recompile Code'

    def calc_enabled(self, action_key):
        return bool(hasattr(self.editor.linked_base.segment, "jumpman_playfield_model") and self.editor.linked_base.segment.assembly_source)

    def perform(self, event):
        self.editor.linked_base.segment.jumpman_playfield_model.compile_assembly_source()


class jumpman_level_list(ViewerActionMixin, segment_select):
    empty_list_name = "No Jumpman Levels Found"

    def is_valid_segment(self, segment):
        state = is_valid_level_segment(segment)
        log.debug(f"is jumpman level {segment}? {state}")
        return state


class JumpmanMouseModeTool(ViewerRadioAction):
    mouse_mode_cls = None

    def calc_name(self, action_key):
        return self.mouse_mode_cls.menu_item_tooltip

    def calc_icon_name(self, action_key):
        return self.mouse_mode_cls.icon

    def calc_enabled(self, action_key):
        return True

    def calc_checked(self, action_key):
        return self.viewer.control.is_mouse_mode(self.mouse_mode_cls)

    def perform(self, action_key):
        self.viewer.control.set_mouse_mode(self.mouse_mode_cls)


class EraseModeMixin:
    def calc_enabled(self, action_key):
        return self.viewer.current_level.is_editing_trigger

class jumpman_select_mode(JumpmanMouseModeTool):
    mouse_mode_cls = jm.AnticDSelectMode

class jumpman_draw_girder_mode(JumpmanMouseModeTool):
    mouse_mode_cls = jm.DrawGirderMode

class jumpman_draw_double_girder_mode(JumpmanMouseModeTool):
    mouse_mode_cls = jm.DrawDoubleGirderMode

class jumpman_draw_ladder_mode(JumpmanMouseModeTool):
    mouse_mode_cls = jm.DrawLadderMode

class jumpman_draw_up_rope_mode(JumpmanMouseModeTool):
    mouse_mode_cls = jm.DrawUpRopeMode

class jumpman_draw_down_rope_mode(JumpmanMouseModeTool):
    mouse_mode_cls = jm.DrawDownRopeMode

class jumpman_erase_girder_mode(EraseModeMixin, JumpmanMouseModeTool):
    mouse_mode_cls = jm.EraseGirderMode

class jumpman_erase_ladder_mode(EraseModeMixin, JumpmanMouseModeTool):
    mouse_mode_cls = jm.EraseLadderMode

class jumpman_erase_rope_mode(EraseModeMixin, JumpmanMouseModeTool):
    mouse_mode_cls = jm.EraseRopeMode

class jumpman_draw_coin_mode(JumpmanMouseModeTool):
    mouse_mode_cls = jm.DrawCoinMode

class jumpman_respawn_mode(JumpmanMouseModeTool):
    mouse_mode_cls = jm.JumpmanRespawnMode



import numpy as np
cimport numpy as np

cdef extern:
    int lib6502_init_cpu(int, int)
    int lib6502_clear_state_arrays(np.uint8_t *buf, np.uint8_t *buf)
    int lib6502_configure_state_arrays(np.uint8_t *buf, np.uint8_t *buf)
    int lib6502_next_frame(np.uint8_t *buf, np.uint8_t *buf, np.uint8_t *buf, np.uint8_t *buf)
    void lib6502_show_next_instruction(np.uint8_t *buf)
    void lib6502_get_current_state(np.uint8_t *buf)
    void lib6502_restore_state(np.uint8_t *buf)
    void lib6502_set_a2_emulation_mode(np.uint8_t value)

def init_emulator(args):
    lib6502_init_cpu(262, 65)  # apple 2 speed

def clear_state_arrays(np.ndarray input not None, np.ndarray output not None):
    cdef np.uint8_t[:] ibuf
    cdef np.uint8_t[:] obuf

    ibuf = input.view(np.uint8)
    obuf = output.view(np.uint8)
    lib6502_clear_state_arrays(&ibuf[0], &obuf[0])

def configure_state_arrays(np.ndarray input not None, np.ndarray output not None):
    cdef np.uint8_t[:] ibuf
    cdef np.uint8_t[:] obuf

    ibuf = input.view(np.uint8)
    obuf = output.view(np.uint8)
    lib6502_configure_state_arrays(&ibuf[0], &obuf[0])

def next_frame(np.ndarray input not None, np.ndarray output not None, np.ndarray breakpoints not None, history_storage):
    cdef np.uint8_t[:] ibuf  # ignored for this emulator
    cdef np.uint8_t[:] obuf
    cdef np.uint8_t[:] dbuf
    cdef np.uint8_t *hbuf
    cdef np.uint8_t[:] tmp
    if history_storage is not None:
        tmp = history_storage.history_array.view(np.uint8)
        hbuf = &tmp[0]
    else:
        hbuf = <np.uint8_t *>0

    ibuf = input.view(np.uint8)
    obuf = output.view(np.uint8)
    dbuf = breakpoints.view(np.uint8)
    bpid = lib6502_next_frame(&ibuf[0], &obuf[0], &dbuf[0], hbuf)
    return bpid

def show_next_instruction(history_storage):
    cdef np.uint8_t *hbuf
    cdef np.uint8_t[:] tmp
    if history_storage is not None:
        tmp = history_storage.history_array.view(np.uint8)
        hbuf = &tmp[0]
    else:
        hbuf = <np.uint8_t *>0
    lib6502_show_next_instruction(hbuf)

def get_current_state(np.ndarray output not None):
    cdef np.uint8_t[:] obuf

    obuf = output.view(np.uint8)
    lib6502_get_current_state(&obuf[0])

def load_disk(int disknum, char *filename, int readonly=0):
    raise NotImplementedError

def restore_state(np.ndarray state not None):
    cdef np.uint8_t[:] sbuf
    sbuf = state.view(np.uint8)
    lib6502_restore_state(&sbuf[0])

def set_a2_emulation_mode(int value):
    lib6502_set_a2_emulation_mode(value)



import numpy as np

from . import SetSelectionCommand, SetRangeCommand

from atrip.disassembler import mini_assemble

import logging
log = logging.getLogger(__name__)


class SetDisasmCommand(SetRangeCommand):
    short_name = "set_disasm_type"
    ui_name = "Set Disassembler Type"
    serialize_order =  [
            ('segment', 'int'),
            ('ranges', 'int_list'),
            ('disasm_type', 'int'),
            ]

    def __init__(self, segment, ranges, disasm_type):
        SetRangeCommand.__init__(self, segment, ranges)
        self.disasm_type = disasm_type

    def get_data(self, orig):
        return self.disasm_type

    def do_change(self, editor, undo):
        indexes = self.range_to_index_function(self.ranges)
        # print(f"{self.short_name}: ranges={self.ranges}, indexes={indexes}")
        undo.flags.index_range = indexes[0], indexes[-1]
        old_data = self.segment.disasm_type[indexes].copy()
        self.segment.disasm_type[indexes] = self.get_data(old_data)
        self.segment.update_data_style_from_disasm_type()
        if self.advance:
            undo.flags.advance_caret_position_in_control = editor.focused_viewer.control
        return old_data

    def undo_change(self, editor, old_data):
        indexes = self.range_to_index_function(self.ranges)
        self.segment.disasm_type[indexes] = old_data
        self.segment.update_data_style_from_disasm_type()


class MiniAssemblerCommand(SetSelectionCommand):
    short_name = "miniasm"
    ui_name = "Assemble"

    def __init__(self, segment, cpu, selection, data, advance=False):
        SetSelectionCommand.__init__(self, segment, selection, data, advance)
        self.cpu = cpu

    def change_data_at_indexes(self, indexes):
        new_data = np.empty(len(indexes) + 100, dtype=np.uint8)
        new_indexes = np.empty(len(indexes) + 100, dtype=np.uint32)
        total = 0
        indexes.sort()  # carets may be out of order, so force increasing
        next_valid_start = 0
        for index in indexes:
            if index < next_valid_start:
                # don't overwrite an instruction in the middle
                continue
            pc = self.segment.origin + index
            d = mini_assemble(self.cpu, self.data, pc)
            count = len(d)
            new_data[total:total + count] = d
            next_valid_start = index + count
            new_indexes[total:total + count] = np.arange(index, next_valid_start)
            total += count
        indexes = new_indexes[0:total]
        old_data = self.segment[indexes]
        self.segment[indexes] = new_data[0:total]
        return old_data, indexes



# Standard library imports.

# Major package imports.
import numpy as np

import sawx.ui.compactgrid as cg

import logging
log = logging.getLogger(__name__)


class SelectionMixin:
    def __str__(self):
        return f"{self.__class__.__name__}: source={self.control.segment_viewer.ui_name} carets=" + " ".join([str(c) for c in self.carets])

    def get_selected_status_message(self):
        carets = list(self.carets_with_selection)
        if len(carets) == 0:
            return ""
        elif len(carets) == 1:
            table = self.table
            viewer = self.control.segment_viewer
            caret = carets[0]
            anchor_start, _ = table.get_index_range(*caret.anchor_start)
            _, anchor_end = table.get_index_range(*caret.anchor_end)
            num = anchor_end - anchor_start
            if num == 1:
                text = f"[1 byte selected ${viewer.get_address_at_index(anchor_start)}]"
            elif num > 0:
                text = f"[{num} bytes selected ${viewer.get_address_at_index(anchor_start)}-${viewer.get_address_at_index(anchor_end)}]"
        else:
            text = "[%d ranges selected]" % (len(carets))
        text += f" from {self.control.segment_viewer.window_title}"
        return text

    def calc_indexes(self, dest_segment, use_carets_without_selections=True):
        table = self.table
        indexes = np.empty(200000, dtype=np.uint32)
        current = 0
        for caret in self.carets:
            index, _ = table.get_index_range(*caret.rc)
            if caret.anchor_start[0] < 0:
                if not use_carets_without_selections:
                    continue
                indexes[current:current + 1] = index
                current += 1
            else:
                anchor_start, _ = table.get_index_range(*caret.anchor_start)
                _, anchor_end = table.get_index_range(*caret.anchor_end)
                count = anchor_end - anchor_start
                indexes[current:current + count] = np.arange(anchor_start, anchor_end, dtype=np.uint32)
                current += count
        dest_indexes = dest_segment.calc_indexes_from_other_segment(indexes[0:current], self.segment)
        return dest_indexes

    def calc_ranges(self, dest_segment, use_carets_without_selections=True):
        """Calculates a list where each entry in the list is a numpy array
        of indexes of each selected range.
        """
        table = self.table
        ranges = []
        current = 0
        for caret in self.carets:
            index, _ = table.get_index_range(*caret.rc)
            if caret.anchor_start[0] < 0:
                if not use_carets_without_selections:
                    continue
                indexes = np.arange(index, index + 1, dtype=np.uint32)
            else:
                anchor_start, _ = table.get_index_range(*caret.anchor_start)
                _, anchor_end = table.get_index_range(*caret.anchor_end)
                indexes = np.arange(anchor_start, anchor_end, dtype=np.uint32)
            dest_indexes = dest_segment.calc_indexes_from_other_segment(indexes, self.segment)
            ranges.append(dest_indexes)
        return ranges

    def convert_carets_from(self, other_char_handler):
        new_carets = []
        table = self.table
        other_table = other_char_handler.table
        other_segment = other_char_handler.segment
        for caret in other_char_handler.carets:
            other_index, _ = other_table.get_index_range(*caret.rc)
            index = self.segment.calc_index_from_other_segment(other_index, other_segment)
            r, c = table.index_to_row_col(index)
            new_caret = cg.Caret(r, c)

            # other_index, _ = other_table.get_index_range(*caret.anchor_start)
            # index = self.segment.calc_index_from_other_segment(other_index, other_segment)
            # r, c = table.index_to_row_col(index)
            # caret.anchor_start = caret.anchor_initial_start = (r, c)

            # _, other_index = other_table.get_index_range(*caret.anchor_end)
            # index = self.segment.calc_index_from_other_segment(other_index, other_segment)
            # r, c = table.index_to_row_col(index)
            # caret.anchor_end = caret.anchor_initial_end = (r, c)

            log.debug(f"converted {caret} from {other_char_handler.control},index={other_index} to {self.control},index={index}, new_caret")
            new_carets.append(new_caret)
        self.carets = new_carets

    def set_caret_from_indexes(self, index, anchor_start=-1, anchor_end=-1):
        table = self.table
        r, c = table.index_to_row_col(index)
        caret = cg.Caret(r, c)
        if anchor_start >= 0:
            r, c = table.index_to_row_col(anchor_start)
            caret.anchor_start = caret.anchor_initial_start = (r, c)
            r, c = table.index_to_row_col(anchor_end)
            caret.anchor_end = caret.anchor_initial_end = (r, c)
        self.carets = [caret]

    def convert_style_to_carets(self):
        table = self.table
        ranges = self.segment.get_style_ranges(selected=True)
        print(ranges)
        new_carets = []
        for start, end in ranges:
            r, c = table.index_to_row_col(end - 1)
            caret = cg.Caret(r, c)
            r, c = table.index_to_row_col(end)
            caret.anchor_end = caret.anchor_initial_end = (r, c)
            r, c = table.index_to_row_col(start)
            caret.anchor_start = caret.anchor_initial_start = (r, c)
            new_carets.append(caret)
        self.carets = new_carets
        print(self.carets)


class FrozenSelection(SelectionMixin, cg.MultiCaretHandler):
    def __init__(self, current_selection):
        cg.MultiCaretHandler.__init__(self)
        self.control = current_selection.control
        self.table = current_selection.table
        self.segment = current_selection.segment
        self.carets = [c.copy() for c in current_selection.carets]


class CurrentSelection(SelectionMixin, cg.MultiCaretHandler):
    def __init__(self, control):
        cg.MultiCaretHandler.__init__(self)
        self.control = control

    @property
    def table(self):
        return self.control.table

    @property
    def segment(self):
        return self.control.segment_viewer.caret_conversion_segment

    def copy(self):
        return FrozenSelection(self)



""" Action definitions for ByteEdit task

"""
import os
import sys

import wx
import wx.lib.dialogs

# traits standins till this whole action code stuff gets replaced
Any = None
Int = None

# from sawx.framework.enthought_api import ActionItem, EditorAction, NameChangeAction, TaskDynamicSubmenuGroup
# import sawx.framework.clipboard as clipboard
from sawx.action import SawxAction

from . import commands
from ..arch.ui.antic_colors import AnticColorDialog
from sawx.ui.dialogs import prompt_for_hex, prompt_for_dec, prompt_for_string, get_file_dialog_wildcard, ListReorderDialog, prompt_for_slice
from ..ui.dialogs import prompt_for_assembler
from ..arch.machine import Machine
from ..byte_edit.commands import SetCommentCommand, ClearCommentCommand, SetLabelCommand, ClearLabelCommand

if sys.platform == "darwin":
    RADIO_STYLE = "toggle"
else:
    RADIO_STYLE = "radio"

import logging
log = logging.getLogger(__name__)


class ViewerAction(SawxAction):
    @property
    def viewer(self):
        return self.task.active_editor.focused_viewer

    @property
    def linked_base(self):
        return self.task.active_editor.focused_viewer.linked_base

    def _update_enabled(self, ui_state):
        if self.enabled_name and ui_state:
            enabled = self._get_attr(self.viewer, self.enabled_name, None)
            if enabled is None:
                enabled = bool(self._get_attr(self.task.active_editor, self.enabled_name, None))
                if enabled is None:
                    log.warning("%s flag does not exist in viewer %s or active editor %s" % (self.enabled_name, self.viewer, self.task_active_editor))
            self.enabled = bool(enabled)
        else:
            self.enabled = True



class FontChoiceGroup(ViewerAction):
    """Dynamic menu group to display the available fonts
    """
    #### 'DynamicSubmenuGroup' interface ######################################

    event_name = 'machine_menu_changed'

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _get_items(self, event_data=None):
        items = []
        if event_data is not None:
            for font in event_data.font_list:
                action = UseFontAction(font=font)
                items.append(ActionItem(action=action))

        return items


class UseFontAction(ViewerAction):
    font = Any
    enabled_name = 'has_font'

    def _name_default(self):
        return "%s" % (self.font['name'])

    def perform(self, event):
        self.viewer.set_font(self.font)


class LoadFontAction(ViewerAction):
    name = 'Load Font...'

    def perform(self, event):
        path = event.task.prompt_local_file_dialog("Load Font")
        if path is not None:
            self.viewer.machine.load_font(event.task, path)

# Assembler syntax section

class AssemblerChoiceGroup(ViewerAction):
    """Dynamic menu group to display the available assemblers used to display
    the disassembly syntax
    """
    #### 'DynamicSubmenuGroup' interface ######################################

    event_name = 'machine_menu_changed'

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _get_items(self, event_data=None):
        items = []
        if event_data is not None:
            for asm in event_data.assembler_list:
                action = UseAssemblerAction(assembler=asm)
                items.append(ActionItem(action=action))

        return items


class UseAssemblerAction(ViewerAction):
    style = RADIO_STYLE
    enabled_name = 'has_cpu'

    assembler = Any

    def _name_default(self):
        return "%s" % (self.assembler['name'])

    def perform(self, event):
        self.linked_base.clear_disassembly(self.viewer.machine)
        self.viewer.machine.set_assembler(self.assembler)

    def _update_checked(self, ui_state):
        self.checked = self.viewer.machine.assembler == self.assembler


class AddNewAssemblerAction(ViewerAction):
    """Add the syntax characteristics for a new assembler.

    """
    name = 'Add New Assembler...'

    def perform(self, event):
        d = prompt_for_assembler(event.task.window.control, "New Assembler")
        if d:
            self.viewer.machine.add_assembler(event.task, d)


class EditAssemblersAction(ViewerAction):
    """Modify the list of assemblers, rearranging the order or editing
    the characteristics of existing assemlers.

    """
    name = 'Edit Assemblers...'

    def perform(self, event):
        items = Machine.assembler_list
        dlg = ListReorderDialog(event.task.window.control, Machine.assembler_list, lambda a:a['name'], prompt_for_assembler, "Manage Assemblers")
        if dlg.ShowModal() == wx.ID_OK:
            asms = dlg.get_items()
            Machine.set_assembler_list(event.task, asms)
            self.viewer.machine.verify_current_assembler()
        dlg.Destroy()


class SetSystemDefaultAssemblerAction(ViewerAction):
    """Mark an assembler that will become the default for future
    editing sessions.

    """
    name = 'Set Current as System Default'

    def perform(self, event):
        Machine.set_system_default_assembler(event.task, self.viewer.machine.assembler)


class FontRendererAction(ViewerAction):
    """This submenu contains a list of all available font renderers. Selecting
    an item in this list will change how the text is displayed in the character
    map window.
    """
    doc_hint = "parent,list"
    submenu_of = "Character Display Modes"
    # Traits
    style = RADIO_STYLE
    enabled_name = 'has_font'

    font_renderer = Any

    def _name_default(self):
        return self.font_renderer.name

    def perform(self, event):
        self.viewer.set_font(font_renderer=self.font_renderer)

    def _update_checked(self, ui_state):
        self.checked = self.viewer.machine.font_renderer == self.font_renderer


class FontMappingAction(ViewerAction):
    """This submenu contains a list of all available character mappings. Most
    platforms use a regular ASCII mapping, but some like the Atari 8-bits have
    different character mappings depending on the usage: the ANTIC mapping if
    looking at screen memory and the ATASCII mapping for normal usage.
    """
    doc_hint = "parent,list"
    submenu_of = "Character Mappings"
    # Traits
    style = RADIO_STYLE
    enabled_name = 'has_font'

    font_mapping = Any

    def _name_default(self):
        return self.font_mapping.name

    def perform(self, event):
        self.viewer.machine.set_font_mapping(self.font_mapping)

    def _update_checked(self, ui_state):
        self.checked = self.viewer.machine.font_mapping == self.font_mapping


class BitmapRendererAction(ViewerAction):
    """This submenu contains a list of all available bitmap renderers.
    Selecting an item in this list will change the rendering of the graphics
    display.
    """
    doc_hint = "parent,list"
    # Traits
    style = RADIO_STYLE
    enabled_name = 'has_bitmap'

    bitmap_renderer = Any

    def _name_default(self):
        return self.bitmap_renderer.name

    def perform(self, event):
        self.viewer.machine.set_bitmap_renderer(self.bitmap_renderer)

    def _update_checked(self, ui_state):
        self.checked = self.viewer.machine.bitmap_renderer == self.bitmap_renderer


class ViewerWidthAction(ViewerAction):
    """Set the number of items per row of the bitmap display. The width can
    mean different things for different viewers (i.e. bitmap widths are in
    byte_values, not pixels), so prompt is based on the viewer.
    """
    name = "Width"
    enabled_name = 'has_width'

    def perform(self, event):
        v = self.viewer
        val = prompt_for_dec(v.control, 'Enter new %s' % v.width_text, 'Set Width', v.width)
        if val is not None and val > 0:
            v.set_width(val)


class ViewerZoomAction(ViewerAction):
    """Set the zoom factor of the viewer, if applicable. This is an integer
    value greater than zero that scales the display size of each item.
    """
    name = "Zoom"
    enabled_name = 'has_zoom'

    def perform(self, event):
        v = self.viewer
        val = prompt_for_dec(v.control, 'Enter new %s' % v.zoom_text, 'Set Zoom', v.zoom)
        if val is not None and val > 0:
            v.set_zoom(val)


class AnticColorAction(ViewerAction):
    """Open a window to choose the color palette from the available colors
    of the ANTIC processor.
    """
    name = 'Use ANTIC Colors...'
    enabled_name = 'has_colors'

    def perform(self, event):
        v = self.viewer
        dlg = AnticColorDialog(v.control, v.machine.antic_color_registers, v.linked_base.cached_preferences)
        if dlg.ShowModal() == wx.ID_OK:
            v.machine.update_colors(dlg.colors)


class UseColorsAction(ViewerAction):
    """Changes the color palette to {name}
    """
    name = 'Use Colors'
    colors = Any
    enabled_name = 'has_colors'

    def perform(self, event):
        self.viewer.machine.update_colors(self.colors)


class ColorStandardAction(ViewerAction):
    """This list sets the color encoding standard for all bitmapped graphics of
    the disk image. Currently supported are:
    """
    doc_hint = "parent,list"
    style = RADIO_STYLE
    enabled_name = 'has_colors'

    color_standard = Int

    def perform(self, event):
        self.viewer.machine.set_color_standard(self.color_standard)

    def _update_checked(self, ui_state):
        self.checked = self.viewer.machine.color_standard == self.color_standard


class TextFontAction(ViewerAction):
    """Open a font selection window to choose the font and size used to display
    the values in the hex grid and the disassembly text.
    """
    name = 'Text Font...'
    enabled_name = 'has_text_font'

    def perform(self, event):
        v = self.viewer
        data = wx.FontData()
        data.EnableEffects(True)
        data.SetColour(v.machine.text_color)
        data.SetInitialFont(v.machine.text_font)
        dlg = wx.FontDialog(self.active_editor.control, data)
        if dlg.ShowModal() == wx.ID_OK:
            data = dlg.GetFontData()
            font = data.GetChosenFont()
            v.machine.set_text_font(font, data.GetColour())
            prefs = self.viewer.linked_base.cached_preferences
            prefs.text_font = font
            self.editor.document.recalc_event(True)


class PredefinedMachineAction(ViewerAction):
    """These are built-in machine definitions that store preset values for
    `Processor`_, `Memory Map`_, `Colors`_, `Font`_, `Character Display`_ and
    `Bitmap Display`_.

    Currently defined machine types are:
    """
    doc_hint = "parent,list"
    # Traits
    style = RADIO_STYLE

    machine = Any

    def _name_default(self):
        return self.machine.name

    def perform(self, event):
        self.viewer.set_machine(self.machine)

    def _update_checked(self, ui_state):
        self.checked = self.viewer.machine == self.machine


class ProcessorTypeAction(ViewerAction):
    """The processor type defines the opcodes displayed in the disassembly
    window and those understood by the mini-assembler.

    Currently supported processors are:
    """
    doc_hint = "parent,list"
    # Traits
    style = RADIO_STYLE
    enabled_name = 'has_cpu'

    disassembler = Any

    def _name_default(self):
        return self.disassembler.name

    def perform(self, event):
        log.debug("setting disassembler to: %s" % self.disassembler.name)
        self.viewer.machine.set_disassembler(self.disassembler)

    def _update_checked(self, ui_state):
        self.checked = self.viewer.machine.disassembler == self.disassembler


class MemoryMapAction(ViewerAction):
    """In the disassembly window, target addresses for jumps, branches, loads,
    stores, compares, etc. can be replaced by text labels that are defined for
    the particular platform. For example, on the Atari 8-bit platform,
    occurrences of ``$E459`` will be replaced by ``SIOV``.

    Some platforms like the Atari 8-bits have hardware locations that perform
    different functions depending on whether it is read from or written to. The
    disassembler can handle this, so reading the location ``$D000`` will show
    the label ``M0PF`` (missile 0 to playfield collisions) while writing will
    show ``HPOSP0`` (set the horizontal position of player 0).

    Currently supported platforms are:
    """
    doc_hint = "parent,list"
    # Traits
    style = RADIO_STYLE
    enabled_name = 'has_cpu'

    memory_map = Any

    def _name_default(self):
        return self.memory_map.name

    def perform(self, event):
        self.viewer.machine.set_memory_map(self.memory_map)

    def _update_checked(self, ui_state):
        self.checked = self.viewer.machine.memory_map == self.memory_map


class IndexRangeAction(ViewerAction):
    """base class for byte modification operations
    """
    doc_hint = "parent,list,level 3"
    parent_doc = """The commands in this menu operate on the current selection to change the byte values to:
    """
    cmd = None
    enabled_name = 'has_editable_bytes'

    def _name_default(self):
        return self.cmd.ui_name

    def get_cmd(self, editor, segment, ranges):
        return self.cmd(segment, ranges)

    def perform(self, event):
        e = self.active_editor
        ranges = self.viewer.control.get_selected_ranges(self.linked_base)
        cmd = self.get_cmd(e, e.segment, ranges)
        e.process_command(cmd)


class ZeroAction(IndexRangeAction):
    """Set to zero"""
    cmd = commands.ZeroCommand
    accelerator = 'Ctrl+0'


class FFAction(IndexRangeAction):
    """Set to $ff"""
    cmd = commands.FFCommand
    accelerator = 'Ctrl+9'


class NOPAction(IndexRangeAction):
    """Set to the NOP command of the current processor"""
    cmd = commands.NOPCommand
    accelerator = 'Ctrl+3'
    enabled_name = 'has_cpu'

    def get_cmd(self, editor, segment, ranges):
        nop = self.viewer.machine.get_nop()
        return self.cmd(segment, ranges, nop)


class SetHighBitAction(IndexRangeAction):
    """Or with $80"""
    cmd = commands.SetHighBitCommand


class ClearHighBitAction(IndexRangeAction):
    """And with $7f"""
    cmd = commands.ClearHighBitCommand


class BitwiseNotAction(IndexRangeAction):
    """Invert every bit in each byte"""
    cmd = commands.BitwiseNotCommand
    accelerator = 'Ctrl+1'


class LeftShiftAction(IndexRangeAction):
    """Shift bits left (multiply by 2), inserting zeros in the low bit"""
    cmd = commands.LeftShiftCommand


class RightShiftAction(IndexRangeAction):
    """Shift bits right (divide by 2), inserting zeros in the high bit"""
    cmd = commands.RightShiftCommand


class LeftRotateAction(IndexRangeAction):
    """Shift bits left where the high bit wraps around to become the new low bit"""
    cmd = commands.LeftRotateCommand
    accelerator = 'Ctrl+<'


class RightRotateAction(IndexRangeAction):
    """Shift bits right where the low bit wraps around to become the new high bit"""
    cmd = commands.RightRotateCommand
    accelerator = 'Ctrl+>'


class ReverseBitsAction(IndexRangeAction):
    """Reverse the bit pattern of each byte; e.g. $c0 or 11000000 in binary becomes 00000011 in binary, $03 in hex"""
    cmd = commands.ReverseBitsCommand


class RandomBytesAction(IndexRangeAction):
    """Generate random bytes to replace the selected data"""
    cmd = commands.RandomBytesCommand


class IndexRangeValueAction(IndexRangeAction):
    prompt = "Enter byte value: (default hex, prefix with # for decimal, %% for binary)"

    def _name_default(self):
        return self.cmd.ui_name + "..."

    def show_dialog(self):
        value = prompt_for_hex(self.viewer.control, self.prompt, self.cmd.ui_name)
        if value is not None:
            cmd = self.cmd(self.linked_base.segment, self.linked_base.carets.selected_ranges, value)
            self.active_editor.process_command(cmd)

    def perform(self, event):
        wx.CallAfter(self.show_dialog)

    def _update_enabled(self, ui_state):
        # Setting a range of values don't make sense for a single byte
        # location, so require it to have an actual range
        return self.viewer.has_editable_bytes and self.linked_base.has_selection


class SetValueAction(IndexRangeValueAction):
    """Prompts the user and sets the data to the specified value"""
    cmd = commands.SetValueCommand


class OrWithAction(IndexRangeValueAction):
    """Logical OR the selected data with the user specified value"""
    cmd = commands.OrWithCommand
    accelerator = 'Ctrl+\\'


class AndWithAction(IndexRangeValueAction):
    """Logical AND the selected data with the user specified value"""
    cmd = commands.AndWithCommand
    accelerator = 'Ctrl+7'


class XorWithAction(IndexRangeValueAction):
    """Logical XOR the selected data with the user specified value"""
    cmd = commands.XorWithCommand
    accelerator = 'Ctrl+6'


class SliceValueAction(IndexRangeValueAction):
    prompt = "Enter byte value: (default hex, prefix with # for decimal, %% for binary)"

    def show_dialog(self):
        slice_obj = prompt_for_slice(self.viewer.control, self.prompt, self.cmd.ui_name)
        if slice_obj is not None:
            cmd = self.cmd(self.linked_base.segment, self.linked_base.carets.selected_ranges, slice_obj)
            self.active_editor.process_command(cmd)


class RampUpAction(SliceValueAction):
    """Starting with the user specified value at the first selected byte, loops
    over each byte in the selection and adds one to the value of the previous
    byte. At $ff, it wraps around to $00.
    """
    cmd = commands.RampUpCommand


class RampDownAction(SliceValueAction):
    """Starting with the user specified value at the first selected byte, loops
    over each byte in the selection and subtracts one from the value of the
    previous byte. At $00, it wraps around to $ff.
    """
    cmd = commands.RampDownCommand


class AddValueAction(IndexRangeValueAction):
    """Adds the user specified value to the data, performing a logical AND with
    $ff if necessary to keep all values in the 8-bit range."""
    cmd = commands.AddValueCommand
    accelerator = 'Ctrl+='


class SubtractValueAction(IndexRangeValueAction):
    """Subtracts the user specified value from the data (AND with $ff if
    necessary). Note the difference between this and `Subtract From`_"""
    cmd = commands.SubtractValueCommand
    accelerator = 'Ctrl+-'


class SubtractFromAction(IndexRangeValueAction):
    """Subtracts the data from the user specified value (AND with $ff if
    necessary). Note the difference between this and `Subtract`_"""
    cmd = commands.SubtractFromCommand
    accelerator = 'Shift+Ctrl+-'


class MultiplyAction(IndexRangeValueAction):
    """Multiply the data from the user specified value (AND with $ff if
    necessary)."""
    cmd = commands.MultiplyCommand
    accelerator = 'Ctrl+8'


class DivideByAction(IndexRangeValueAction):
    """Divides the data by the user specified value by the data, ignoring the
    remainder. Note the difference between this and `Divide From`_"""
    cmd = commands.DivideByCommand
    accelerator = 'Ctrl+/'


class DivideFromAction(IndexRangeValueAction):
    """Divides the data from the user specified value (that is to say: dividing
    the user specified value by the data), ignoring the remainder. Note the
    difference between this and `Divide By`_"""
    cmd = commands.DivideFromCommand
    accelerator = 'Shift+Ctrl+/'


class ReverseSelectionAction(IndexRangeAction):
    """Reverses the order of bytes in the selection"""
    cmd = commands.ReverseSelectionCommand


class ReverseGroupAction(IndexRangeValueAction):
    prompt = "Enter number of bytes in each group: (default hex, prefix with # for decimal, %% for binary)"
    cmd = commands.ReverseGroupCommand


class StartTraceAction(ViewerAction):
    name = "Start New Disassembly Trace"
    accelerator = 'F12'
    enabled_name = 'has_cpu'

    def perform(self, event):
        self.viewer.start_trace()


class AddTraceStartPointAction(ViewerAction):
    name = "Add Trace Start Point"
    accelerator = 'F11'
    enabled_name = 'has_cpu'
    tooltip = 'Start a trace at the caret or at all instructions in the selected ranges'

    def perform(self, event):
        v = self.viewer
        # check if selected a range:
        ranges, indexes = v.get_selected_ranges_and_indexes()
        s = v.segment
        checked = set()
        for i in indexes:
            pc = v.control.table.lines.get_instruction_start_pc(i + s.origin)
            if pc not in checked:
                v.trace_disassembly(pc)
                checked.add(pc)
        v.linked_base.force_refresh()


class ApplyTraceSegmentAction(ViewerAction):
    name = "Apply Trace to Segment"
    accelerator = 'F10'
    tooltip = 'Copy the results of the trace to the current segment'
    enabled_name = 'has_cpu'

    def perform(self, event):
        cmd = commands.ApplyTraceSegmentCommand(self.viewer.segment)
        self.active_editor.process_command(cmd)


class ClearTraceAction(ViewerAction):
    name = "Clear Trace"
    accelerator = 'F9'
    tooltip = 'Clear the current trace'
    enabled_name = 'has_cpu'

    def perform(self, event):
        cmd = commands.ClearTraceCommand(self.active_editor.document.collection)
        self.active_editor.process_command(cmd)


class CopyAsCBytesAction(ViewerAction):
    """Copy the current selection as text where each byte is converted to the
    C source code representation.
    """
    name = 'Copy as C Data'
    enabled_name = 'can_copy'

    def perform(self, event):
        v = self.viewer
        ranges, indexes = v.get_selected_ranges_and_indexes()
        data = v.segment[indexes]
        text = ",\n".join([",".join(["0x%02x" % d for d in c]) for c in [data[i:i+8] for i in range(0, len(data), 8)]])
        data_obj = wx.TextDataObject()
        data_obj.SetText(text)
        clipboard.set_clipboard_object(data_obj)


class CopyAsReprAction(ViewerAction):
    """Copy the current selection as a text string containing a string (with
    escaped characters where necessary) that reproduces the bytes in the
    selection.

    Python note: both double quotes and single quotes are escaped as hex values
    so the resulting string is safe to use inside either of those characters as
    the string delimiter.
    """
    name = 'Copy as Escaped String'
    enabled_name = 'can_copy'

    def perform(self, event):
        v = self.viewer
        ranges, indexes = v.get_selected_ranges_and_indexes()
        data = v.segment[indexes]
        s1 = repr(data.tobytes())
        q = s1[0]
        text = s1[1:-1]  # remove leading/trailing quotes
        if q == "'":
            # double quotes are literal, single quotes are escaped
            text = text.replace('"', "\\x22").replace("\\'", "\\x27")
        else:
            # single quotes are literal, double are escaped
            text = text.replace("'", "\\x27").replace('\\"', "\\x22")
        data_obj = wx.TextDataObject()
        data_obj.SetText(text)
        clipboard.set_clipboard_object(data_obj)


def prompt_for_comment(e, s, ranges, desc):
    existing = s.get_first_comment(ranges)
    text = prompt_for_string(e.window.control, desc, "Add Comment", existing)
    if text is not None:
        cmd = SetCommentCommand(s, ranges, text)
        e.process_command(cmd)


class AddCommentAction(ViewerAction):
    """Add a text comment to a byte location.

    A comment is associated with a single byte, so although a range can be
    selected, the comment is applied to only the first byte in the range.

    Bytes with comments will be highlighted in all displays.
    """
    name = 'Add Comment'
    accelerator = 'Alt+C'

    def is_range(self, event):
        return self.active_editor.can_copy

    def get_index(self, event):
        return self.viewer.linked_base.carets.current.index

    def perform(self, event):
        e = self.active_editor
        s = e.segment
        if self.is_range(event):
            ranges = s.get_style_ranges(selected=True)
            if len(ranges) == 1:
                desc = "Enter comment for first byte of range:\n%s" % e.get_label_of_first_byte(ranges)
            else:
                desc = "Enter comment for first byte of each range:\n%s" % e.get_label_of_first_byte(ranges)
        else:
            ranges = []
        if not ranges:
            index = self.get_index(event)
            ranges = [(index, index+1)]
            desc = "Enter comment for location %s" % index
        prompt_for_comment(e, s, ranges, desc)


class AddCommentPopupAction(AddCommentAction):
    name = 'Add Comment'

    def is_range(self, event):
        return event.popup_data["in_selection"]

    def get_index(self, event):
        return event.popup_data["index"]


class RemoveCommentAction(ViewerAction):
    """Remove any comments that are in the selected range, or if no selection
    from the current caret position.

    """
    name = 'Remove Comment'
    accelerator = 'Shift+Alt+C'

    def perform(self, event):
        e = self.active_editor
        s = e.segment
        if e.can_copy:
            ranges = s.get_style_ranges(selected=True)
        else:
            index = e.caret_index
            ranges = [(index, index+1)]
        if ranges:
            cmd = ClearCommentCommand(s, ranges)
            e.process_command(cmd)


class RemoveCommentPopupAction(ViewerAction):
    name = 'Remove Comment'

    def perform(self, event):
        e = self.active_editor
        s = e.segment
        if event.popup_data["in_selection"]:
            ranges = s.get_style_ranges(selected=True)
        else:
            index = event.popup_data["index"]
            ranges = [(index, index+1)]
        if ranges:
            cmd = ClearCommentCommand(s, ranges)
            e.process_command(cmd)


def prompt_for_label(e, s, addr, desc):
    existing = s.memory_map.get(addr, "")
    text = prompt_for_string(e.window.control, desc, "Add Label", existing)
    if text is not None:
        cmd = SetLabelCommand(s, addr, text)
        e.process_command(cmd)


class AddLabelAction(ViewerAction):
    """Add a label to a byte location.

    Like `Add Comment`_, a label is associated with a single byte, so although
    a range can be selected, the comment is applied to only the first byte in
    the range.

    Unlike comments, labels are *not* highlighted and are only shown in the
    disassembly window.
    """
    name = 'Add Label'
    accelerator = 'Alt+L'
    enabled_name = 'has_origin'

    def get_ranges(self, editor, segment, event):
        if editor.can_copy:  # has selected ranges
            ranges = segment.get_style_ranges(selected=True)
        else:
            ranges = [(editor.caret_index, editor.caret_index + 1)]
        return ranges

    def process_ranges(self, editor, segment, ranges):
        index = ranges[0][0]
        addr = index + segment.origin
        return addr

    def process(self, editor, segment, addr):
        desc = "Enter label for address $%04x" % addr
        prompt_for_label(editor, segment, addr, desc)

    def perform(self, event):
        e = self.active_editor
        s = e.segment
        ranges = self.get_ranges(e, s, event)
        addr = self.process_ranges(e, s, ranges)
        self.process(e, s, addr)


class AddLabelPopupAction(AddLabelAction):
    def get_ranges(self, editor, segment, event):
        if event.popup_data["in_selection"]:
            ranges = segment.get_style_ranges(selected=True)
        else:
            index = event.popup_data["index"]
            ranges = [(index, index+1)]
        return ranges


class RemoveLabelAction(AddLabelAction):
    """Remove the label at the current caret position, or if there is a
    selection, all labels in the selected range.

    """
    name = 'Remove Label'
    accelerator = 'Shift+Alt+L'

    def process_ranges(self, editor, segment, ranges):
        return ranges

    def process(self, editor, segment, ranges):
        cmd = ClearLabelCommand(segment, ranges)
        editor.process_command(cmd)


class RemoveLabelPopupAction(RemoveLabelAction):
    def get_ranges(self, editor, segment, event):
        if event.popup_data["in_selection"]:
            ranges = segment.get_style_ranges(selected=True)
        else:
            index = event.popup_data["index"]
            ranges = [(index, index+1)]
        return ranges



# color blending utilities

import math

import numpy as np


def get_blended_color_registers(colors, blend_color):
    registers = []
    base_blend = [(r * 7)/8 for r in blend_color]
    for c in colors:
        r = [c[i]/8 + base_blend[i] for i in range(3)]
        registers.append(r)
    return registers


def get_dimmed_color_registers(colors, background_color, dimmed_color):
    registers = []
    dimmed_difference = [b - d for b, d in zip(background_color, dimmed_color)]
    for c in colors:
        r = [max(0, c[i]- dimmed_difference[i]) for i in range(3)]
        registers.append(r)
    return registers


def calc_blended_rgb(rgb_colors, blend_color):
    registers = np.zeros((256, 3), dtype=np.uint8)
    base_blend = [(r * 7) for r in blend_color]
    for i in range(len(registers)):
        registers[i] = [(int(rgb_colors[i,j]) + base_blend[j]) // 8 for j in range(3)]
    return registers


def calc_dimmed_rgb(rgb_colors, background_color, dimmed_color):
    registers = np.zeros((256, 3), dtype=np.uint8)
    dimmed_difference = [b - d for b, d in zip(background_color, dimmed_color)]
    for i in range(len(registers)):
        registers[i] = [max(0, int(rgb_colors[i,j]) - dimmed_difference[j]) for j in range(3)]
    return registers



import os
import sys

import numpy as np

import wx

from atrip import Container, Segment

from atrip.disassembler import flags, dd

from sawx.ui import compactgrid as cg
from ..editors.linked_base import VirtualTableLinkedBase

from ..ui.segment_grid import SegmentGridControl

from ..viewer import SegmentViewer
from .emulator import EmulatorViewerMixin

import logging
log = logging.getLogger(__name__)


KFEST_HACK = False


class InstructionHistoryTable(cg.VirtualTable):
    column_labels = ["^Instruction", "^Result"]
    column_sizes = [21, 12]

    def __init__(self, linked_base):
        self.virtual_linked_base = linked_base
        s = linked_base.segment
        self.current_num_rows = 0
        self.history_entries = None
        self.visible_history_start_row = 0
        self.visible_history_lookup_table = None
        cg.VirtualTable.__init__(self, len(self.column_labels), s.origin)
        self.selected_rows = set()

    @property
    def emulator(self):
        return self.virtual_linked_base.emulator

    def calc_num_rows(self):
        return self.current_num_rows

    def calc_last_valid_index(self):
        return self.current_num_rows * self.items_per_row

    def get_label_at_index(self, index):
        row = (index // self.items_per_row) - self.visible_history_start_row
        if row < 0:
            return "----"
        try:
            emu = self.emulator
            return "%04x" % (emu.cpu_history[row][0])
        except IndexError:
            return "----"

    def get_row_label_text(self, start_line, num_lines, step=1):
        last_line = min(start_line + num_lines, self.num_rows)
        emu = self.emulator
        for line in range(start_line, last_line, step):
            h = emu.cpu_history[line]
            t = h['disassembler_type']
            if t == flags.DISASM_NEXT_INSTRUCTION:
                h = h.view(dtype=dd.HISTORY_BREAKPOINT_DTYPE)
                t = h['disassembler_type_cpu']
            if t == flags.DISASM_ATARI800_HISTORY or t == flags.DISASM_6502_HISTORY or t == flags.DISASM_ATARI800_VBI_START or t == flags.DISASM_ATARI800_VBI_END or t == flags.DISASM_ATARI800_DLI_START or t == flags.DISASM_ATARI800_DLI_END:
                h = h.view(dtype=dd.HISTORY_ATARI800_DTYPE)
                x = h['tv_cycle']
                y = h['tv_line']
                if x & 0x80:
                    x = x & 0x7f
                    y = y | 0x100
                yield "%3d %3d" % (y, x)
            elif t == flags.DISASM_FRAME_START or t == flags.DISASM_FRAME_END:
                h = h.view(dtype=dd.HISTORY_FRAME_DTYPE)
                f = h['frame_number']
                yield "f%d" % (f)
            else:
                yield "%d" % (emu.cpu_history[line][0])

    def find_previous_line(self, start, flag_type):
        emu = self.emulator
        line = start
        while line > 0:
            line -= 1
            h = emu.cpu_history[line]
            t = h['disassembler_type']
            if flag_type == t:
                break
        return line

    def find_next_line(self, start, flag_type):
        emu = self.emulator
        line = start
        while line < self.num_rows - 1:
            line += 1
            h = emu.cpu_history[line]
            t = h['disassembler_type']
            if flag_type == t:
                break
        return line

    def find_frame_instruction(self, line):
        line = self.find_previous_line(line, flags.DISASM_FRAME_END)
        return line

    def calc_row_label_width(self, view_params):
        return view_params.calc_text_width("256 256")

    def get_value_style(self, row, col):
        t = self.parsed
        if t is None:
            return "", 0
        try:
            text = t[row - self.visible_history_start_row][col]
        except IndexError:
            print(f"tried row {row} out of {self.visible_history_lookup_table}")
            text = f"row {row} out of bounds"
        if row in self.selected_rows:
            style = cg.selected_bit_mask
        else:
            style = 0
        return text, style

    def clear_selected_style(self):
        # self.style[:] &= (0xff ^ selected_bit_mask)
        self.selected_rows = set()

    def set_selected_index_range(self, index1, index2):
        # self.style[index1:index2] |= selected_bit_mask
        row1 = index1 // self.items_per_row
        row2 = index2 // self.items_per_row
        for r in range(row1, row2 + 1):
            self.selected_rows.add(r)
            print(f"adding {r} to selected_rows")

    def prepare_for_drawing(self, start_row, visible_rows, start_cell, visible_cells):
        emu = self.emulator
        self.visible_history_start_row = start_row
        self.parsed = emu.calc_stringified_history(start_row, visible_rows)

    @property
    def needs_rebuild(self):
        emu = self.emulator
        return not self.current_num_rows == emu.num_cpu_history_entries

    def rebuild(self):
        v = self.virtual_linked_base
        emu = v.emulator
        self.current_num_rows = len(emu.cpu_history)
        c = Container(emu.cpu_history.entries.view(np.uint8), force_numpy_data=True)
        v.segment = Segment(c, 0)
        # print("CPU HISTORY ENTRIES", self.current_num_rows)
        self.init_boundaries()

        # for now, clear selection when anything changes so we don't have to
        # track the line numbers as they should get moved up the screen as more
        # instructions get added to the bottom
        self.clear_selected_style()


class InstructionHistoryGridControl(SegmentGridControl):
    default_table_cls = InstructionHistoryTable

    extra_keybinding_desc = {
        "caret_move_frame_down": "Shift+Pagedown",
        "caret_move_frame_up": "Shift+Pageup",
        "caret_move_frame_top": "Shift+Left",
    }
    keybinding_desc = SegmentGridControl.keybinding_desc
    keybinding_desc.update(extra_keybinding_desc)

    if KFEST_HACK:
        kfest_detach_update = False

    def calc_default_table(self, linked_base):
        table = self.default_table_cls(linked_base)
        table.rebuild()  # find number of rows so scrollbars can be set properly
        return table

    def calc_line_renderer(self):
        return cg.VirtualTableLineRenderer(self, 2, widths=self.default_table_cls.column_sizes, col_labels=self.default_table_cls.column_labels)

    def move_viewport_to_bottom(self):
        self.move_viewport_origin((self.table.current_num_rows, 0))

    def caret_move_frame_down(self, evt, f):
        row = self.caret_handler.current.rc[0]
        row = self.table.find_next_line(row, flags.DISASM_FRAME_END)
        self.caret_handler.move_carets_to(row, 0)
        self.caret_handler.validate_carets()

    def caret_move_frame_up(self, evt, f):
        row = self.caret_handler.current.rc[0]
        row = self.table.find_previous_line(row, flags.DISASM_FRAME_END)
        self.caret_handler.move_carets_to(row, 0)
        self.caret_handler.validate_carets()

    def caret_move_frame_top(self, evt, f):
        self.caret_handler.move_carets_to(0, 0)
        self.caret_handler.validate_carets()

    def handle_select_start(self, evt, row, col, flags):
        doc = self.segment_viewer.document
        print("LEFT DOWN!", row, col, doc, doc.emulator_running)
        if KFEST_HACK:
            if doc.emulator_paused:
                try:
                    self.kfest_detach_update = True
                    emu = self.table.emulator
                    h = emu.cpu_history[row]
                    t = h['disassembler_type']
                    frame_number = emu.kfest_history_to_frame_number[row]
                    print(f"history: {row}: {frame_number}, {emu.kfest_frame_number_to_history[frame_number]}")
                    print(f"history: {row}: {h}")
                    emu.restore_restart(0, frame_number - 1)
                    emu.kfest_step_history(frame_number, row)
                    doc.emulator_update_screen_event(True)
                    doc.priority_level_refresh_event(100)
                except IndexError:
                    pass
                except KeyError:
                    pass
            else:
                self.kfest_detach_update = False

        return super().handle_select_start(evt, row, col, flags)

    def recalc_view(self):
        if KFEST_HACK:
            if self.kfest_detach_update:
                log.critical("KFEST: DETACHED")
                return
        log.debug(f"recalc_view: {self}")
        self.table.rebuild()
        cg.CompactGrid.recalc_view(self)
        self.move_viewport_to_bottom()

    def refresh_view(self):
        if self.IsShown():
            log.debug("refreshing %s" % self)
            if self.table.needs_rebuild:
                self.recalc_view()
            else:
                SegmentGridControl.refresh_view(self)
        else:
            log.debug("skipping refresh of hidden %s" % self)


class InstructionHistoryViewer(EmulatorViewerMixin, SegmentViewer):
    name = "cpuhistory"

    ui_name = "Instruction History"

    viewer_category = "Emulator"

    control_cls = InstructionHistoryGridControl

    priority_refresh_frame_count = 10

    # initialization

    @classmethod
    def replace_linked_base(cls, linked_base):
        # the new linked base decouples the cursor here from the other segments
        c = Container(np.arange(40, dtype=np.uint8))
        segment = Segment(c, 0)
        return VirtualTableLinkedBase(editor=linked_base.editor, segment=segment)

    def create_post(self):
        self.linked_base.table = self.control.table

    # properties

    @property
    def table(self):
        return self.control.table

    # @on_trait_change('linked_base.editor.document.byte_values_changed')
    def byte_values_changed(self, index_range):
        log.debug("byte_values_changed: %s index_range=%s" % (self, str(index_range)))
        if index_range is not Undefined:
            self.table.rebuild()

    # @on_trait_change('linked_base.editor.document.byte_style_changed')
    def byte_style_changed(self, index_range):
        log.debug("byte_style_changed: %s index_range=%s" % (self, str(index_range)))
        if index_range is not Undefined:
            self.table.rebuild()

    def recalc_data_model(self):
        self.table.rebuild()

    def do_priority_level_refresh(self):
        self.control.recalc_view()
        self.refresh_view(True)



import os
import time
import inspect
import pkg_resources

from . import errors

from atrip.memory_map import size_codes

import logging
log = logging.getLogger(__name__)


class DataInfo:
    def __init__(self, label, addr, type_code, count):
        self.label = label
        self.addr = addr
        self.type_code = type_code
        self.count = count
        self.byte_count = self.count * (size_codes[type_code] + 1)

    def __str__(self):
        return f"DataInfo: {self.label}@${self.addr:04x}, {self.count}x{self.type_code}"

    def __repr__(self):
        return f"DataInfo: {self.label}@${self.addr:04x}, {self.count}x{self.type_code}"


class AssemblerResult:
    def __init__(self, verbose=False):
        self.verbose = verbose
        self.timestamp = time.ctime()
        self.errors = []
        self.segments = []
        self.transitory_equates = {}
        self.equates = {}
        self.labels = {}
        self.data_info = {}
        self.addr_to_label = {}
        self.addr_type_code = {}
        self.first_addr = None
        self.last_addr = None
        self.current_bytes = []
        self.user_variable_labels = None

    def __bool__(self):
        return not bool(self.errors)

    def add_label(self, label, addr):
        label = label.lower()
        addr = int(addr, 16)
        self.labels[label] = addr
        if addr not in self.addr_to_label:
            # first label takes priority if multiple labels for same addr
            self.addr_to_label[addr] = label

    def generate_data_info(self):
        assembled_addrs = sorted(self.addr_type_code.keys())
        print(assembled_addrs)
        print(self.addr_to_label)
        if len(assembled_addrs) < 1:
            return

        index = 0
        self.data_info = {}
        current_info = None

        def add_data(info):
            if self.verbose: print(f"adding data info: {info}")
            self.data_info[info.label] = info

        try:
            while True:
                if current_info is None:
                    # find label
                    while True:
                        start = assembled_addrs[index]
                        try:
                            label = self.addr_to_label[start]
                            current_type_code = self.addr_type_code[start]
                            if current_type_code == "code":
                                raise KeyError(f"address type_code {current_type_code} is not data")
                            current_info = DataInfo(label, start, current_type_code, 1)
                        except KeyError as e:
                            print(f"index={index}: start={start:x}, {e}")
                            index += 1
                        else:
                            break

                # when we get here, we have a label that is pointing to some
                # data, so find next label, gap in addresses assembled, or
                # change in data type_code
                if self.verbose: print(f"found data info: {current_info}")
                last = start
                while True:
                    index += 1
                    print(f"trying {index:x}")
                    end = assembled_addrs[index]
                    print(f"found {end:x} at {index}")
                    if end > start + current_info.count + 1:
                        # gap in addresses, meaning the data block has ended
                        add_data(current_info)
                        current_info = None
                        break
                    elif end in self.addr_to_label or self.addr_type_code[end] != current_type_code:
                        add_data(current_info)
                        current_info = None
                        break
                    else:
                        current_info.count += 1
        except IndexError:
            # no more addresses in assembled_addrs array
            if self.verbose: print(f"finished processing assembled_addrs")
            if current_info is not None:
                add_data(current_info)
        if self.verbose: print(self.data_info)

    def find_next_address_with_label(self, addr):
        for a in sorted(self.addr_to_label):
            if a > addr:
                break
        else:
            raise KeyError
        return a


class Assembler:
    name = "<base>"
    ui_name = "<pretty name>"
    cpu = "<cpu>"

    comment_char = ";"
    origin = "*="
    data_byte = ".byte"
    data_byte_prefix = "$"
    data_byte_separator = ", "

    def __init__(self, verbose=False):
        self.verbose = verbose

    def assemble(self, source):
        result = AssemblerResult()
        result.errors = ["No assemblers found"]
        return result


_assemblers = None

_default_assembler = None

def _find_assemblers():
    assemblers = []
    for entry_point in pkg_resources.iter_entry_points('atrip.assemblers'):
        mod = entry_point.load()
        log.debug(f"find_assembler: Found module {entry_point.name}={mod.__name__}")
        for name, obj in inspect.getmembers(mod):
            if inspect.isclass(obj) and Assembler in obj.__mro__[1:]:
                log.debug(f"find_assemblers:   found assembler class {name}")
                assemblers.append(obj)
    return assemblers

def find_assemblers():
    global _assemblers

    if _assemblers is None:
        _assemblers = _find_assemblers()
    return _assemblers

def find_assembler_by_name(name):
    items = find_assemblers()
    for c in items:
        if c.text_type == name:
            return c()
    raise KeyError(f"Unknown assembler {name}")

def set_default_assembler(name):
    global _default_assembler

    # let KeyError propagate up to indicate unknown assembler
    asm = find_assembler_by_name(name)
    _default_assembler = asm

def get_default_assembler():
    global _default_assembler

    if _default_assembler is None:
        find_assemblers()
        if len(_assemblers) == 0:
            _default_assembler = Assembler()
        elif len(_assemblers) == 1:
            _default_assembler = _assemblers[0]()
    return _default_assembler



import re
import bisect

import numpy as np
import numpy as np

from sawx.utils.command import Command, UndoInfo

from omnivore.utils.searchutil import AlgorithmSearcher

import logging
log = logging.getLogger(__name__)


class FindAllCommand(Command):
    short_name = "find"
    ui_name = "Find"

    def __init__(self, start_caret_index, search_text, error, repeat=False, reverse=False):
        Command.__init__(self)
        self.start_caret_index = start_caret_index
        self.current_caret_index = start_caret_index
        self.search_text = search_text
        self.error = error
        self.repeat = repeat
        self.reverse = reverse
        self.current_match_index = -1
        self.origin = -1
        self.search_copy = None

    def __str__(self):
        return "%s %s" % (self.ui_name, repr(self.search_text))

    def get_search_string(self):
        return bytes.fromhex(self.search_text)

    def get_searchers(self, editor):
        return editor.searchers

    def perform(self, editor, undo):
        self.origin = editor.segment.origin
        self.search_copy = editor.segment.tobytes()
        self.all_matches = []
        self.match_ids = {}
        undo.flags.changed_document = False
        if self.error:
            undo.flags.message = self.error
        else:
            errors = []
            match_dict = {}
            editor.segment.clear_style_bits(match=True)
            searchers = self.get_searchers(editor)
            log.debug(f"Using searchers {searchers}")
            for searcher_cls in searchers:
                try:
                    searcher = searcher_cls(editor, self.search_text, self.search_copy)
                    for start, end in searcher.matches:
                        if start in self.match_ids:
                            if searcher.ui_name not in self.match_ids[start]:
                                self.match_ids[start] +=", %s" % searcher.ui_name
                            if end > match_dict[start]:
                                match_dict[start] = end
                        else:
                            self.match_ids[start] = searcher.ui_name
                            match_dict[start] = end
                except ValueError as e:
                    errors.append(str(e))

            if errors:
                undo.flags.message = " ".join(errors)
            else:
                self.all_matches = [(start, match_dict[start]) for start in sorted(match_dict.keys())]

                #print "Find:", self.all_matches
                if len(self.all_matches) == 0:
                    undo.flags.message = "Not found"
                else:
                    editor.segment.set_style_ranges(self.all_matches, match=True)
                    # Need to use a tuple in order for bisect to search the
                    # list of tuples
                    caret_tuple = (self.current_caret_index, 0)
                    self.current_match_index = bisect.bisect_left(self.all_matches, caret_tuple)
                    if self.current_match_index >= len(self.all_matches):
                        self.current_match_index = 0
                    match = self.all_matches[self.current_match_index]
                    start = match[0]
                    log.debug("Starting at match_index %d = %s" % (self.current_match_index, match))
                    control = editor.focused_viewer.control
                    control.caret_handler.set_caret_from_indexes(start, match[0], match[1])
                    undo.flags.sync_caret_from_control = control
                    undo.flags.message = ("Match %d of %d, found at $%04x in %s" % (self.current_match_index + 1, len(self.all_matches), start + self.origin, self.match_ids[start]))
            undo.flags.refresh_needed = True


class FindNextCommand(Command):
    short_name = "findnext"
    ui_name = "Find Next"

    def __init__(self, search_command):
        Command.__init__(self)
        self.search_command = search_command

    def get_index(self, editor):
        cmd = self.search_command
        caret_tuple = (cmd.current_caret_index, 0)
        match_index = bisect.bisect_right(cmd.all_matches, caret_tuple)
        if match_index == cmd.current_match_index:
            match_index += 1
        if match_index >= len(cmd.all_matches):
            match_index = 0
        cmd.current_match_index = match_index
        return match_index

    def perform(self, editor, undo):
        cmd = self.search_command
        undo.flags.changed_document = False
        index = self.get_index(editor)
        all_matches = cmd.all_matches
        #print "FindNext:", all_matches
        try:
            match = all_matches[index]
            start = match[0]
            control = editor.focused_viewer.control
            control.caret_handler.set_caret_from_indexes(start, match[0], match[1])
            undo.flags.sync_caret_from_control = control
            c = self.search_command
            undo.flags.message = ("Match %d of %d, found at $%04x in %s" % (index + 1, len(all_matches), start + c.origin, c.match_ids[start]))
            cmd.current_caret_index = start
        except IndexError:
            pass
        undo.flags.refresh_needed = True


class FindPrevCommand(FindNextCommand):
    short_name = "findprev"
    ui_name = "Find Previous"

    def get_index(self, editor):
        cmd = self.search_command
        caret_tuple = (cmd.current_caret_index, 0)
        match_index = bisect.bisect_left(cmd.all_matches, caret_tuple)
        match_index -= 1
        if match_index < 0:
            match_index = len(cmd.all_matches) - 1
        cmd.current_match_index = match_index
        return match_index


class FindAlgorithmCommand(FindAllCommand):
    short_name = "findalgorithm"
    ui_name = "Find using expression"

    def get_searchers(self, editor):
        return [AlgorithmSearcher]



import numpy as np

from . import ChangeStyleCommand

import logging
log = logging.getLogger(__name__)


class SetCommentCommand(ChangeStyleCommand):
    short_name = "set_comment"
    ui_name = "Comment"
    serialize_order =  [
            ('segment', 'int'),
            ('ranges', 'int_list'),
            ('text', 'string'),
            ]

    def __init__(self, segment, ranges, text):
        ChangeStyleCommand.__init__(self, segment)
        # Only use the first byte of each range
        self.ranges = self.convert_ranges(ranges)
        log.debug("%s operating on ranges: %s" % (self.ui_name, str(ranges)))
        self.text = text
        # indexes = ranges_to_indexes(self.ranges)
        # self.index_range = indexes[0], indexes[-1]
        # if len(ranges) == 1:
        #     self.ui_name = "%s @ %04x" % (self.ui_name, self.segment.origin + indexes[0])

    def __str__(self):
        if len(self.text) > 20:
            text = self.text[:20] + "..."
        else:
            text = self.text
        return "%s: %s" % (self.ui_name, text)

    def convert_ranges(self, ranges):
        return tuple([(items[0], items[0] + 1) for items in ranges])

    def set_undo_flags(self, flags):
        flags.byte_style_changed = True
        # flags.index_range = self.index_range

    def clamp_ranges_and_indexes(self, editor):
        return self.ranges, None

    def change_comments(self, ranges, indexes):
        self.segment.set_comment_ranges(ranges, self.text)

    def do_change(self, editor, undo):
        ranges, indexes = self.clamp_ranges_and_indexes(editor)
        old_data = self.segment.get_comment_restore_data(ranges)
        self.change_comments(ranges, indexes)
        return old_data

    def undo_change(self, editor, old_data):
        self.segment.restore_comments(old_data)


class ClearCommentCommand(SetCommentCommand):
    short_name = "clear_comment"
    ui_name = "Remove Comment"

    def __init__(self, segment, ranges):
        SetCommentCommand.__init__(self, segment, ranges, "")

    def convert_ranges(self, ranges):
        # When clearing comments, we want to look at every space, not just the
        # first byte of each range like setting comments
        return ranges

    def change_comments(self, ranges, indexes):
        self.segment.clear_comment_ranges(ranges)


class PasteCommentsCommand(ClearCommentCommand):
    short_name = "paste_comments"
    ui_name = "Paste Comments"
    serialize_order =  [
            ('segment', 'int'),
            ('ranges', 'int_list'),
            ('cursor', 'int'),
            ('bytes', 'string'),
            ]

    def __init__(self, segment, ranges, cursor, bytes, *args, **kwargs):
        # remove zero-length ranges
        r = [(start, end) for start, end in ranges if start != end]
        ranges = r
        if not ranges:
            # use the range from cursor to end
            ranges = [(cursor, len(segment))]
        ClearCommentCommand.__init__(self, segment, ranges, bytes)
        self.cursor = cursor
        self.comments = bytes.tostring().splitlines()
        self.num_lines = len(self.comments)

    def __str__(self):
        return "%s: %d line%s" % (self.ui_name, self.num_lines, "" if self.num_lines == 1 else "s")

    def clamp_ranges_and_indexes(self, editor):
        disasm = editor.disassembly.table.disassembler
        count = self.num_lines
        comment_indexes = []
        clamped = []
        for start, end in self.ranges:
            index = start
            log.debug("starting range %d:%d" % (start, end))
            while index < end and count > 0:
                comment_indexes.append(index)
                pc = index + self.segment.origin
                log.debug("comment at %d, %04x" % (index, pc))
                try:
                    index = disasm.get_next_instruction_pc(pc)
                    count -= 1
                except IndexError:
                    count = 0
            clamped.append((start, index))
            if count <= 0:
                break
        return clamped, comment_indexes

    def change_comments(self, ranges, indexes):
        """Add comment lines as long as we don't go out of range (if specified)
        or until the end of the segment or the comment list is exhausted.

        Depends on a valid disassembly to find the lines; we are adding a
        comment for the first byte in each statement.
        """
        log.debug("ranges: %s" % str(ranges))
        log.debug("indexes: %s" % str(indexes))
        self.segment.set_comments_at_indexes(ranges, indexes, self.comments)
        self.segment.set_style_at_indexes(indexes, comment=True)



# Debugger breakpoint definition
import numpy as np

MAIN_MEMORY_SIZE = 1<<16

FRAME_STATUS_DTYPE = np.dtype([
    ("cycles_since_power_on", np.uint64),
    ("instructions_since_power_on", np.uint64),
    ("cycles_user", np.uint64),
    ("instructions_user", np.uint64),
    ("frame_number", np.uint32),
    ("current_cycle_in_frame", np.uint32),
    ("final_cycle_in_frame", np.uint32),
    ("current_instruction_in_frame", np.uint32),
    ("scan_lines_since_power_on", np.uint32),
    ("unused0", np.uint32, 3),

    ("breakpoint_id", np.int16),
    ("current_scan_line_in_frame", np.int16),
    ("unused1", np.uint16, 2),

    ("frame_status", np.uint8),
    ("use_memory_access", np.uint8),
    ("brk_into_debugger", np.uint8),
    ("unused2", np.uint8, 5),

    ("unused3", np.uint64, 6), # fill header to 128 bytes

    ("memory_access", np.uint8, MAIN_MEMORY_SIZE),
    ("access_type", np.uint8, MAIN_MEMORY_SIZE),
])

ACCESS_TYPE_READ = 1
ACCESS_TYPE_WRITE = 2
ACCESS_TYPE_EXECUTE = 4
ACCESS_TYPE_VIDEO = 8
ACCESS_TYPE_DISPLAY_LIST = 16

default_access_type_colors = {
    0: (0, 0, 0),
    ACCESS_TYPE_READ: (0, 255, 0),
    ACCESS_TYPE_WRITE: (255, 0, 0),
    ACCESS_TYPE_EXECUTE: (255, 255, 0),
    ACCESS_TYPE_VIDEO: (0, 0, 255),
    ACCESS_TYPE_DISPLAY_LIST: (255, 0, 255),
}


NUM_BREAKPOINT_ENTRIES = 256
TOKENS_PER_BREAKPOINT = 64
TOKEN_LIST_SIZE = (NUM_BREAKPOINT_ENTRIES * TOKENS_PER_BREAKPOINT)

# Breakpoint #0 is used internally for stepping the CPU
DEBUGGER_COMMANDS_DTYPE = np.dtype([
    ("num_breakpoints", np.uint32),
    ("last_pc", np.int32),
    ("unused", np.uint32, 14),
    ("reference_value", np.int64, NUM_BREAKPOINT_ENTRIES),
    ("breakpoint_type", np.uint8, NUM_BREAKPOINT_ENTRIES),
    ("breakpoint_status", np.uint8, NUM_BREAKPOINT_ENTRIES),
    ("tokens", np.uint16, TOKEN_LIST_SIZE),
])

# Breakpoints are address of PC to break at before executing code at that
# address. Breakpoints are only evaluated if the corresponding
# `breakpoint_status` value is BREAKPOINT_ENABLED. `num_breakpoints` serves as
# a speed optimization: the `breakpoints` list will not be scanned past that
# number of entries. There are a maximum of 256 breakpoint entries.

# Watchpoints use the same status values as breakpoints, but as they are
# variable length include an index into the watchpoints list and the length in
# words in `watchpoint_index` and `watchpoint_length` respectively.
# `num_watchpoints` serves the same purpose as `num_breakpoints`. There are a
# maximum of 256 watchpoint entries. The list of words pointed to by
# `watchpoint_index` are mathematical terms in postfix notation, the evaluation
# of which yields a 16 bit unsigned value, interpreted as a true or false
# condition whether the watchpoint should be triggered. True is a non-zero
# value and false is zero.

# Breakpoint/watchpoint status values. For watchpoints, the status value can be
# changed by the watchpoint processor if it finds a problem in a rule. Note
# that watchpoints are only processed when their status is BREAKPOINT_ENABLED,
# so it is up to the calling program to change the state back to this after an
# error is corrected.

BREAKPOINT_EMPTY = 0
BREAKPOINT_ENABLED = 0x20
BREAKPOINT_DISABLED = 0x40
BREAKPOINT_ERROR = 0x80
EVALUATION_ERROR = 0x81  # a problem with the postfix definition
STACK_UNDERFLOW = 0x82  # too many operators/not enough values
STACK_OVERFLOW = 0x83  # too many values

# breakpoint types
BREAKPOINT_CONDITIONAL = 0
BREAKPOINT_COUNT_INSTRUCTIONS = 0x1
BREAKPOINT_COUNT_CYCLES = 0x2
BREAKPOINT_AT_RETURN = 0x3
BREAKPOINT_COUNT_FRAMES = 0x4
BREAKPOINT_INFINITE_LOOP = 0x5
BREAKPOINT_BRK_INSTRUCTION = 0x6
BREAKPOINT_PAUSE_AT_FRAME_START = 0x7
BREAKPOINT_COUNT_LINES = 0x8


# contitional breakpoint definitions

OP_UNARY = 0x1000
OP_BINARY = 0x2000
VALUE_ARGUMENT = 0x3000

OP_MASK = 0xf000
TOKEN_MASK = 0x0fff

# operations
END_OF_LIST = 0
OP_BITWISE_AND = (102 | OP_BINARY)
OP_BITWISE_NOT = (103 | OP_UNARY)
OP_BITWISE_OR = (104 | OP_BINARY)
OP_DIV = (105 | OP_BINARY)
OP_EQ = (106 | OP_BINARY)
OP_EXP = (107 | OP_BINARY)
OP_GE = (108 | OP_BINARY)
OP_GT = (109 | OP_BINARY)
OP_LE = (110 | OP_BINARY)
OP_LOGICAL_AND = (111 | OP_BINARY)
OP_LOGICAL_NOT = (112 | OP_UNARY)
OP_LOGICAL_OR = (113 | OP_BINARY)
OP_LSHIFT = (114 | OP_BINARY)
OP_LT = (115 | OP_BINARY)
OP_MINUS = (116 | OP_BINARY)
OP_MULT = (117 | OP_BINARY)
OP_NE = (118 | OP_BINARY)
OP_PLUS = (119 | OP_BINARY)
OP_RSHIFT = (120 | OP_BINARY)
OP_UMINUS = (121 | OP_UNARY)
OP_UPLUS = (122 | OP_UNARY)
REG_A = (201)
REG_X = (202)
REG_Y = (203)
REG_S = (204)
REG_N = (205)
REG_V = (206)
REG_B = (207)
REG_D = (208)
REG_I = (209)
REG_Z = (210)
REG_C = (211)
REG_PC = (212)
REG_SP = REG_S
EMU_SCANLINE = (213)
EMU_COLOR_CLOCK = (214)
EMU_VBI_START = (215)  # transition to VBI
EMU_IN_VBI = (216)  # inside VBI
EMU_VBI_END = (217)  # transition out of VBI
EMU_DLI_START = (218)  # transition to DLI
EMU_IN_DLI = (219)  # inside DLI
EMU_DLI_END = (220)  # transition out of DLI
NUMBER = (301 | VALUE_ARGUMENT)
OPCODE_TYPE = (302 | VALUE_ARGUMENT)

COUNT_INSTRUCTIONS = (401 | VALUE_ARGUMENT)
COUNT_CYCLES = (402 | VALUE_ARGUMENT)

OPCODE_READ = 1
OPCODE_WRITE = 2
OPCODE_RETURN = 4
OPCODE_INTERRUPT = 8



import wx
import numpy as np
import jsonpickle

from sawx.errors import ClipboardError
from sawx.clipboard import calc_composite_object

from atrip import Segment
from atrip.stringifier import find_stringifier_by_name

import logging
log = logging.getLogger(__name__)


# Full list of valid data formats:
#
# >>> import wx
# >>> [x for x in dir(wx) if x.startswith("DF_")]
# ['DF_BITMAP', 'DF_DIB', 'DF_DIF', 'DF_ENHMETAFILE', 'DF_FILENAME', 'DF_HTML',
# 'DF_INVALID', 'DF_LOCALE', 'DF_MAX', 'DF_METAFILE', 'DF_OEMTEXT',
# 'DF_PALETTE', 'DF_PENDATA', 'DF_PRIVATE', 'DF_RIFF', 'DF_SYLK', 'DF_TEXT',
# 'DF_TIFF', 'DF_UNICODETEXT', 'DF_WAVE']


def get_data_object_by_format(data_obj, fmt):
    # First try a composite object, then simple: have to handle both
    # cases
    try:
        d = data_obj.GetObject(fmt)
    except AttributeError:
        d = data_obj
    return d


def get_data_object_value(data_obj, name):
    # First try a composite object, then simple: have to handle both
    # cases
    try:
        fmt = data_obj.GetFormat()
    except AttributeError:
        fmt = data_obj.GetPreferredFormat()
    if fmt.GetId() == name:
        d = get_data_object_by_format(data_obj, fmt)
        return d.GetData().tobytes()
    raise ClipboardError("Expecting %s data object, found %s" % (name, fmt.GetId()))


def create_numpy_clipboard_blob(ranges, indexes, segment, control):
    print(ranges)
    rects = control.get_rects_from_selections()
    if rects:
        blob = NumpyRectBlob(ranges, indexes, rects, segment, control)
    else:
        if len(ranges) == 1:
            blob = NumpyBlob(ranges, indexes, segment, control)
        elif len(ranges) > 1:
            blob = NumpyMultipleBlob(ranges, indexes, segment, control)
    return blob


class ClipboardBlob:
    data_format_name = "_base_"
    ui_name = "_base_"

    def __init__(self):
        self.data = None
        self.indexes = None
        self.style = None
        self.relative_comment_indexes = None
        self.comments = None
        self.num_rows = None
        self.num_cols = None
        self.num_regions = 1
        self.dest_items_per_row = None
        self.dest_carets = None

    @property
    def serialized(self):
        return jsonpickle.encode(self).encode('utf-8')

    @property
    def data_obj(self):
        data_obj = wx.CustomDataObject(self.data_format_name)
        data_obj.SetData(self.serialized)
        return data_obj

    def text_data_obj(self, stringifier="hexify"):
        try:
            s = find_stringifier_by_name(stringifier)
            log.warning(f"Stringifier {stringifier} not found; using default")
        except KeyError:
            s = find_stringifier_by_name("hexify")
        text = s.calc_text(self.data)
        text_obj = wx.TextDataObject()
        text_obj.SetText(text)
        return text_obj


class TextBlob(ClipboardBlob):
    """Supports unpacking data objects only."""
    data_format_name = "text"
    ui_name = "Text"

    def __init__(self, data_obj):
        super().__init__()
        self.unpack_data_object(data_obj)

    def unpack_data_object(self, data_obj):
        fmts = data_obj.GetAllFormats()
        if wx.DF_TEXT in fmts:
            value = data_obj.GetText().encode('utf-8')
        elif wx.DF_UNICODETEXT in fmts:  # for windows
            value = data_obj.GetText().encode('utf-8')
        else:
            raise ClipboardError("Unsupported format type for %s: %s" % (self.data_format_name, ", ".join([str(f) in fmts])))
        self.data = np.fromstring(value, dtype=np.uint8)


class JumpmanBlob(ClipboardBlob):
    data_format_name = "jumpman,objects"
    ui_name = "Jumpman Objects"

    def __init__(self, objects):
        super().__init__()
        self.data = objects


class NumpyBlob(ClipboardBlob):
    data_format_name = "numpy"
    ui_name = "Single Selection"

    def __init__(self, ranges, indexes, segment, control):
        super().__init__()
        self.ranges = ranges
        self.num_regions = len(self.ranges)
        self.indexes = indexes
        self.get_data_from_segment(segment, control)
        self.style, self.relative_comment_indexes, self.comments = segment.calc_selected_index_metadata(indexes)

    def get_data_from_segment(self, segment, control):
        r = self.ranges[0]
        self.data = segment[r[0]:r[1]]


class NumpyMultipleBlob(NumpyBlob):
    data_format_name = "numpy,multiple"
    ui_name = "Multiple Selection"

    def get_data_from_segment(self, segment, control):
        self.data = segment[self.indexes]


class NumpyRectBlob(NumpyMultipleBlob):
    def __init__(self, ranges, indexes, rects, segment, control):
        super().__init__(ranges, indexes, segment, control)
        self.get_rects_from_segment(rects, segment, control)

    def get_rects_from_segment(self, rects, segment, control):
        r = rects[0]  # FIXME: handle multiple rects
        self.num_rows, self.num_cols, self.data = control.get_data_from_rect(r)


def parse_data_obj(data_obj, viewer):
    if wx.DF_TEXT in data_obj.GetAllFormats() or wx.DF_UNICODETEXT in data_obj.GetAllFormats():  # for windows
        blob = TextBlob(data_obj)
    else:
        fmt = data_obj.GetPreferredFormat()
        name = fmt.GetId()
        value = get_data_object_value(data_obj, name)
        try:
            blob = jsonpickle.decode(value.decode('utf-8'))
            print("DECODED BLOB!", blob)
        except IndexError:
            raise ClipboardError(f"Failed unpacking data_obj {name}: {value}")
    blob.dest_carets = viewer.control.caret_handler.copy()
    return blob



import zlib
import io

import numpy as np

from .. import errors
from ..compressor import Compressor


class ZLibCompressor(Compressor):
    """NOTE: this is the GNU zip compression, not unix compress"""
    compression_algorithm = "zlib"

    def calc_unpacked_data(self, byte_data):
        try:
            unpacked = zlib.decompress(bytes(byte_data))
        except zlib.error as e:
            raise errors.InvalidAlgorithm(e)
        return unpacked

    def calc_packed_data(self, byte_data, media=None):
        try:
            packed = zlib.compress(bytes(byte_data), 9)
        except zlib.error as e:
            raise errors.InvalidAlgorithm(e)
        return packed



# Debugger breakpoint definition
import numpy as np


HISTORY_ENTRY_DTYPE = np.dtype([
    ("pc", np.uint16),
    ("target_addr", np.uint16),
    ("num_bytes", np.uint8),
    ("disassembler_type", np.uint8),
    ("flag", np.uint8),
    ("cycles", np.uint8),
    ("instruction", np.uint8, 16),
])

HISTORY_6502_DTYPE = np.dtype([
    ("pc", np.uint16),
    ("target_addr", np.uint16),
    ("num_bytes", np.uint8),
    ("disassembler_type", np.uint8),
    ("flag", np.uint8),
    ("cycles", np.uint8),
    ("instruction", np.uint8, 3),
    ("a", np.uint8),
    ("x", np.uint8),
    ("y", np.uint8),
    ("sp", np.uint8),
    ("sr", np.uint8),
    ("before1", np.uint8),
    ("after1", np.uint8),
    ("before2", np.uint8),
    ("after2", np.uint8),
    ("before3", np.uint8),
    ("after3", np.uint8),
    ("tv_cycle", np.uint8),
    ("tv_line", np.uint8),
])

HISTORY_ATARI800_DTYPE = np.dtype([
    ("pc", np.uint16),
    ("target_addr", np.uint16),
    ("num_bytes", np.uint8),
    ("disassembler_type", np.uint8),
    ("flag", np.uint8),
    ("cycles", np.uint8),
    ("instruction", np.uint8, 3),
    ("a", np.uint8),
    ("x", np.uint8),
    ("y", np.uint8),
    ("sp", np.uint8),
    ("sr", np.uint8),
    ("before1", np.uint8),
    ("after1", np.uint8),
    ("before2", np.uint8),
    ("after2", np.uint8),
    ("before3", np.uint8),
    ("after3", np.uint8),
    ("tv_cycle", np.uint8),
    ("tv_line", np.uint8),
])

HISTORY_FRAME_DTYPE = np.dtype([
    ("frame_number", np.uint32),
    ("num_bytes", np.uint8),
    ("disassembler_type", np.uint8),
    ("flag", np.uint8),
    ("cycles", np.uint8),
    ("instruction", np.uint8, 16),
])

HISTORY_INTERRUPT_DTYPE = np.dtype([
    ("pc", np.uint16),
    ("target_addr", np.uint16),
    ("num_bytes", np.uint8),
    ("disassembler_type", np.uint8),
    ("flag", np.uint8),
    ("cycles", np.uint8),
    ("instruction", np.uint8, 16),
])

HISTORY_BREAKPOINT_DTYPE = np.dtype([
    ("pc", np.uint16),
    ("breakpoint_id", np.uint8),
    ("breakpoint_type", np.uint8),
    ("num_bytes", np.uint8),
    ("disassembler_type", np.uint8),
    ("flag", np.uint8),
    ("disassembler_type_cpu", np.uint8),
    ("instruction", np.uint8, 16),
])

EMULATOR_HISTORY_HEADER_DTYPE = np.dtype([
    ("num_allocated_entries", np.int32),
    ("num_entries", np.int32),
    ("first_entry_index", np.int32),
    ("latest_entry_index", np.int32),
    ("cumulative_count", np.uint32),
])



import os
import tempfile

import numpy as np

from .persistence import Serializable

import logging
log = logging.getLogger(__name__)


class RestartTree(Serializable):
    name = None

    serializable_attributes = ['restarts']

    def __init__(self):
        self.emulator_start = Restart(0, None)
        self.restarts = [self.emulator_start]

    #### dunder methods

    def __len__(self):
        return len(self.restarts)

    def __iter__(self):
        for r in self.restarts:
            yield r

    def __getitem__(self, index):
        try:
            index = int(index)
        except TypeError:
            try:
                return [self.restarts[i] for i in index]
            except TypeError:
                raise TypeError("argument must be a slice or an integer")
        else:
            if index < 0:
                index += len( self )
            return self.restarts[index]  # will raise IndexError here

    def create_restart(self, restart_number, frame_number):
        restart = self.restarts[restart_number]
        parent = restart.get_restart(frame_number)
        index = len(self.restarts)
        new_restart = Restart(index, parent, frame_number)
        self.restarts.append(new_restart)
        return new_restart

    def get_summary(self):
        # Return list of tuples for use in checkpoint tree. Each tuple in in form (parent restart number, start frame, restart number, last frame)
        restarts = [(-1, 0, 0, self.emulator_start.end_frame)]
        for r in self.restarts[1:]:
            restarts.append((r.parent.restart_number, r.start_frame, r.restart_number, r.end_frame))
        return restarts


class Restart(Serializable):
    name = None

    serializable_attributes = ['frame_history']
    serializable_computed = {'frame_history'}

    def __init__(self, restart_number, parent, start_frame=0):
        self.restart_number = restart_number
        self.parent = parent
        self.start_frame = start_frame
        self.end_frame = start_frame
        self.frame_history = self.calc_history_iterable()

    def calc_history_iterable(self):
        return dict()

    ##### Serialization

    def calc_computed_attribute(self, key):
        if key == 'frame_history':
            return [list(a) for a in self.frame_history.items()]
        return getattr(self, key).copy()

    def restore_computed_attributes(self, state):
        self.frame_history = self.calc_history_iterable()
        for frame_number, data in state['frame_history']:
            self.frame_history[frame_number] = data

    ##### Storage indexes

    def __len__(self):
        return self.end_frame

    # def __iter__(self):
    #     keys = sorted(self.frame_history.keys())
    #     for k in keys:
    #         yield self.frame_history[k]

    def keys(self):
        return sorted(self.frame_history.keys())

    def is_memorable(self, frame_number):
        # return frame_number % 10 == 0
        return True

    def get_previous_frame(self, frame_cursor):
        n = frame_cursor - 1
        while n > 0:
            if n in self.frame_history:
                return n
            n -= 1
        raise IndexError("No previous frame")

    def get_next_frame(self, frame_cursor):
        n = frame_cursor + 1
        largest = max(self.frame_history.keys())
        while n < largest:
            if n in self.frame_history:
                return n
            n += 1
        raise IndexError("No next frame")

    ##### Storage

    def save_frame(self, frame_number, data):
        # History is saved in a big list, which will waste space for empty
        # entries but makes things extremely easy to manage. Simply delete
        # a history entry by setting it to NONE.
        frame_number = int(frame_number)
        self.frame_history[frame_number] = data
        self.end_frame = frame_number

    ##### Retrieval

    def __getitem__(self, index):
        try:
            index = int(index)
        except TypeError:
            try:
                return [self.get_frame(i) for i in index]
            except TypeError:
                raise TypeError("argument must be a slice or an integer")
        else:
            if index < 0:
                index += len( self )
            return self.get_frame(index)  # will raise IndexError here

    def get_restart(self, frame_number):
        frame_number = int(frame_number)
        if frame_number < self.start_frame:
            parent = self.parent
            if parent is None:
                raise IndexError("Initial start of emulator has no parent")
            else:
                restart = parent.get_restart(frame_number)
        else:
            restart = self
        return restart

    def get_frame(self, frame_number):
        parent = self.get_restart(frame_number)  # could raise IndexError
        frame_number = int(frame_number)
        return self.frame_history[frame_number]

    ##### Compact

    def decimate(self):
        """Remove old history items according to an algorithm that discards
        some portion of the older history as time goes on
        """
        pass



import importlib
import functools

import numpy as np

from .disassemblers import dtypes as dd
from .disassemblers import flags

import logging
log = logging.getLogger(__name__)


try:
    from .disassemblers.miniasm import MiniAssembler, processors, get_miniasm, mini_assemble
    from .disassemblers.disasm import DisassemblyConfig, ParsedDisassembly
    from .disassemblers.history import HistoryStorage, StringifiedHistory
    from .disassemblers import libudis, cputables
except RuntimeError:
    log.warning("cputables.py not generated; disassembler and mini assembler will not be available")
except ModuleNotFoundError:
    log.warning("libudis C extensions not created; disassembler and mini assembler will not be availabe")
    raise
except ImportError as e:
    log.warning(f"libudis C extension not loaded (likely an undefined symbol):\n{str(e)}")

try:
    from .disassemblers.valid_cpus import valid_cpu_ids, cpu_id_to_name, cpu_name_to_id
except:
    valid_cpu_ids = [10]
    cpu_name_to_id = {"6502": 10}
    cpu_id_to_name = {10: "6502"}



def create_history_dtype(num_entries, history_dtype):
    size = (num_entries + 1) * dd.HISTORY_ENTRY_DTYPE.itemsize
    dtype_entries = list(dd.EMULATOR_HISTORY_HEADER_DTYPE.descr)
    dtype_entries.append(("entries", history_dtype, num_entries))
    print(f"new dtype: {dtype_entries}")
    return np.dtype(dtype_entries)

def create_history(num_entries, history_dtype):
    dtype = create_history_dtype(num_entries, history_dtype)
    # print(f"new dtype: {dtype.itemsize} bytes")
    history = np.zeros(1, dtype=dtype)
    history['num_allocated_entries'] = num_entries
    clear_history(history)
    return history

def clear_history(history):
    history['num_entries'] = 0
    history['first_entry_index'] = 0
    history['latest_entry_index'] = -1



import os
import sys
from collections import namedtuple

import numpy as np
import wx

from ..ui.screen import BitmapScreen

from sawx.ui import compactgrid as cg
from sawx.ui import checkpointtree as ct
from sawx.utils.command import DisplayFlags
from ..ui import segment_grid as sg
from ..viewer import SegmentViewer
from .info import VirtualTableInfoViewer
from ..editors.linked_base import VirtualTableLinkedBase

import logging
log = logging.getLogger(__name__)


class EmulatorViewerMixin:
    viewer_category = "Emulator"

    has_caret = False

    @property
    def emulator(self):
        return self.linked_base.editor.document.emulator

    @property
    def can_copy(self):
        return False

    @property
    def linked_base_segment_identifier(self):
        return ""

    def use_default_view_params(self):
        pass

    def restore_view_params(self, params):
        pass

    def update_toolbar(self):
        pass

    def recalc_data_model(self):
        # self.control.recalc_view()
        self.control.Refresh()

    def recalc_view(self):
        # self.control.recalc_view()
        self.control.Refresh()

    def update_window_title(self):
        self.update_caption()
        # FIXME: probably shouldn't know this much about the internals to call
        # blah.blah.title_bar
        self.control.GetParent().title_bar.Refresh()


class VideoViewer(EmulatorViewerMixin, SegmentViewer):
    name = "video"

    ui_name = "Emulator Video Output"

    has_caret = False

    # # effectively don't allow emulator screen viewers to use the priority
    # # refresh; they will always be explicitly refreshed with the
    # # emulator_update_screen_event
    # priority_refresh_frame_count = 10000000

    def set_event_handlers(self):
        # override SegmentViewer handlers, which skips the priority level
        # refresh
        self.document.emulator_breakpoint_event += self.on_emulator_breakpoint
        self.document.emulator_update_screen_event += self.on_emulator_update_screen

    @classmethod
    def create_control(cls, parent, linked_base, mdict):
        return BitmapScreen(parent, linked_base.emulator)

    def show_caret(self, control, index, bit):
        pass

    @property
    def window_title(self):
        emu = self.emulator
        return f"{emu.ui_name} (frame {emu.current_frame_number})"

    #### event handlers

    def on_emulator_update_screen(self, evt):
        log.debug("process_emulator_update_screen for %s using %s; flags=%s" % (self.control, self.linked_base, str(evt)))
        self.do_emulator_update_screen()

    def do_emulator_update_screen(self):
        self.control.show_frame(force=True)
        self.update_window_title()

    ##### Spring Tab interface

    def get_notification_count(self):
        return 0


class CheckpointViewer(EmulatorViewerMixin, SegmentViewer):
    name = "checkpoint"

    ui_name = "Checkpoint Viewer"

    has_caret = False

    priority_refresh_frame_count = 59 # about once per second

    def set_event_handlers(self):
        SegmentViewer.set_event_handlers(self)  # skip viewer mixin handlers, which skips the priority level refresh
        self.document.emulator_breakpoint_event += self.on_emulator_breakpoint
        self.document.emulator_update_screen_event += self.on_emulator_update_screen
        self.control.Bind(ct.EVT_CHECKPOINT_SELECTED, self.on_selected)
        self.control.Bind(ct.EVT_CHECKPOINT_RANGE_SELECTED, self.on_range_selected)

    @classmethod
    def create_control(cls, parent, linked_base, mdict):
        return ct.CheckpointTree(parent, linked_base.emulator)

    def show_caret(self, control, index, bit):
        pass

    @property
    def window_title(self):
        return "Emulator Checkpoints"

    def recalc_view(self):
        self.control.recalc_view()

    #### event handlers

    def on_selected(self, evt):
        print("RESTART SELECTED!", evt.GetRestartNumber(), evt.GetFrameNumber(), evt.GetLine())
        doc = self.document
        doc.pause_emulator()
        doc.checkpoint_restore(evt.GetRestartNumber(), evt.GetFrameNumber())
        self.editor.selected_checkpoint_range = None

    def on_range_selected(self, evt):
        print(f"RANGE SELECTED! {evt.start.frame_number}@{evt.start.restart_number} -> {evt.end.frame_number}@{evt.end.restart_number}")
        doc = self.document
        doc.pause_emulator()
        doc.checkpoint_restore(evt.end.restart_number, evt.end.frame_number)
        self.editor.selected_checkpoint_range = (evt.start, evt.end)

    def on_emulator_update_screen(self, evt):
        log.debug("process_emulator_update_screen for %s using %s; flags=%s" % (self.control, self.linked_base, str(evt)))
        self.do_emulator_update_screen()

    def do_emulator_update_screen(self):
        self.control.recalc_view()
        self.control.refresh_view()


class CPU6502Table(sg.SegmentVirtualTable):
    want_col_header = True
    want_row_header = False

    col_labels = ["A", "X", "Y", "N", "V", "-", "B", "D", "I", "Z", "C", "SP", "PC"]
    col_sizes = [2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 2, 4]
    p_bit = [0, 0, 0, 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01, 0]

    # cpu state is not necessarily in "A,X,Y,SP..." order, so use the dtype
    # names of the cpu state array to look up which positions they should be
    # mapped to. The array below is the default from atari800 which is
    # a, p, sp, x, y, _, pc. It's overridden below in compute_col_lookup
    col_from_cpu_state = [0, 3, 4, 1, 1, 1, 1, 1, 1, 1, 1, 2, 6]

    def calc_last_valid_index(self):
        return self.items_per_row

    def get_data_style_view(self, linked_base):
        data = linked_base.emulator.cpu_state
        self.compute_col_lookup(data.dtype)
        return data, None

    def compute_col_lookup(self, dtype):
        label_lookup = {label:i for i,label in enumerate(dtype.names)}
        col_from_cpu_state = [0] * len(self.col_labels)
        for i, d in enumerate(self.col_labels):
            if d in "NV-BDIZC":
                d = "P"
            if d in label_lookup:
                col_from_cpu_state[i] = label_lookup[d]
        self.col_from_cpu_state = col_from_cpu_state

    def get_value_style(self, row, col):
        val = self.data[self.col_from_cpu_state[col]]
        if col < 3 or col == 11:
            text = "%02x" % val
        elif col < 11:
            text = self.col_labels[col] if val & self.p_bit[col] else "-"
        else:
            text = "%04x" % val
        return text, self.style[col]

    def get_label_at_index(self, index):
        return "cpu"


class CPUParamTableViewer(EmulatorViewerMixin, VirtualTableInfoViewer):
    name = ""

    viewer_category = "Emulator"

    ui_name = "<pretty name>"

    control_cls = sg.SegmentVirtualGridControl

    @property
    def linked_base_segment_identifier(self):
        return ""

    def recalc_view(self):
        self.control.recalc_view()


class CPU6502Viewer(CPUParamTableViewer):
    name = "cpu6502"

    ui_name = "6502 CPU Registers"

    override_table_cls = CPU6502Table


class DtypeTable(sg.SegmentVirtualTable):
    emulator_dtype_name = "<from subclass>"
    col_labels = ["hex", "dec", "bin"]
    col_sizes = [4, 4, 8]
    want_col_header = False
    want_row_header = True
    verbose = False

    def calc_num_rows(self):
        self.row_labels = []
        offsets = []
        self.label_char_width = 1
        for i, label in enumerate(self.data.dtype.names):
            if label.startswith("_") and not self.verbose:
                continue
            self.row_labels.append(label)
            offsets.append(i)
            size = len(label)
            if size > self.label_char_width:
                self.label_char_width = size
        self.row_offsets_into_data = np.asarray(offsets, dtype=np.int32)
        return len(self.row_labels)

    def calc_last_valid_index(self):
        return (self.items_per_row * self.num_rows) - 1

    def get_data_style_view(self, linked_base):
        data = linked_base.emulator.calc_dtype_data(self.emulator_dtype_name)
        return data, None

    def create_row_labels(self):
        # already generated above
        pass

    def get_row_label_text(self, start_line, num_lines, step=1):
        last_line = min(start_line + num_lines, self.num_rows)
        for line in range(start_line, last_line, step):
            yield self.row_labels[line]

    def calc_row_label_width(self, view_params):
        return max([view_params.calc_text_width(r) for r in self.row_labels])

    def get_value_style(self, row, col):
        val = self.data[self.row_offsets_into_data[row]]
        try:
            if col == 0:
                text = f"{val:02x}"
            elif col == 1:
                text = f"{val}"
            elif col == 2:
                text = f"{val:08b}"
            index, _ = self.get_index_range(row, col)
            style = self.style[index]
        except TypeError:
            text = "..."
            style = 0
        return text, style

    def get_label_at_index(self, index):
        return self.emulator_dtype_name


# This dynamically creates clases based on the emulator-defined dtypes.
#
# Each resulting class looks like this ANTIC example, except for everywhere
# that ANTIC appears, the other class name is substituted. Much repetition! So,
# instead, we create the classes in a loop.
#
# class ANTICTableViewer(CPUParamTableViewer):
#     name = "antic"
#     ui_name = "ANTIC Registers"
#     override_table_cls = type('ANTICTable', (DtypeTable,), {'emulator_dtype_name': 'ANTIC'})

for dtype_name in ['ANTIC', 'GTIA', 'POKEY', 'PIA']:
    clsname = '%sViewer' % dtype_name
    cls = type(clsname, (CPUParamTableViewer,), {
        'name': dtype_name.lower(),
        'ui_name': '%s Registers' % dtype_name,
        'override_table_cls': type('%sTable' % dtype_name, (DtypeTable,), {'emulator_dtype_name': dtype_name}),
        })
    globals()[clsname] = cls



#!/usr/bin/env python

import wx

import numpy as np


data = [
    (-1, 0, 0, 350),  # root
    (0, 60, 1, 220),  # 1
    (0, 60, 2, 140),
    (0, 60, 3, 180),
    (0, 100, 4, 140),
    (0, 100, 5, 140),  # 5
    (0, 100, 6, 320),
    (0, 200, 7, 240),
    (1, 180, 8, 260),
    (3, 140, 9, 240),
    (0, 200, 10, 260), # 10
    (1, 200, 11, 240),
    (11, 220, 12, 300),
    (11, 220, 13, 320),
    (3, 100, 14, 260),
]


def get_order(source, parent_level):
    """Find restarts with given parent level.

    Building the list from bottom right up, so first restart at the largest
    frame number is first, then later restarts at the same frame number after
    that. Checkpoints that occur at earlier frames then are added, proceeding
    down to frame zero.
    """
    subset_lookup = {}
    for entry in [d for d in source if d[0] == parent_level]:
        start = entry[1]
        if start not in subset_lookup:
            subset_lookup[start] = []
        subset_lookup[start].append(entry)
    processing_order = []
    for items in reversed(list(subset_lookup.values())):
        for item in items:
            processing_order.append(item)
            processing_order.extend(get_order(source, item[2]))
    return processing_order


class RestartLine:
    def __init__(self, data, level):
        self.start_frame = data[1]
        self.end_frame = data[3]
        self.parent = data[0]
        self.restart_number = data[2]
        try:
            current_largest_level = max(level[self.start_frame:self.end_frame + 1])
        except ValueError:
            current_largest_level = 0
        self.level = current_largest_level + 1
        level[self.start_frame:self.end_frame + 1] = self.level
        self.interesting_frames = {
            self.start_frame: "Checkpoint from run {self.parent}",
            self.end_frame: "Last frame",
        }

    def __repr__(self):
        return f"{self.restart_number}: {self.start_frame}->{self.end_frame} @ {self.level}"

    def is_frame_in_line(self, frame):
        return frame >= self.start_frame and frame <= self.end_frame


class RestartLines:
    def __init__(self, processing_order):
        self.lines = [None] * len(processing_order)
        self.last_frame = max([d[3] for d in processing_order])
        self.highest_level = 0
        self.generate_lines(processing_order)

    def __str__(self):
        txt_lines = []
        for line in self.lines:
            txt_lines.append(repr(line))
        return "\n".join(txt_lines)

    def __iter__(self):
        for line in self.lines:
            yield line

    def __getitem__(self, index):
        return self.lines[index]

    def generate_lines(self, processing_order):
        level = np.zeros(self.last_frame, dtype=np.int16) - 1
        for d in processing_order:
            line = RestartLine(d, level)
            self.lines[line.restart_number] = line
            if line.level > self.highest_level:
                self.highest_level = line.level
            #print(level)
            # print("generate", line, line.level)

    def find_line(self, frame_number, level):
        for line in self.lines:
            if frame_number >= line.start_frame and frame_number <= line.end_frame and level == line.level:
                return line
        return None

    def find_parent(self, child_line):
        return self.lines[child_line.parent]


wxEVT_CHECKPOINT_RESTORE = wx.NewEventType()
EVT_CHECKPOINT_RESTORE = wx.PyEventBinder(wxEVT_CHECKPOINT_RESTORE, 1)
wxEVT_CHECKPOINT_HOVER = wx.NewEventType()
EVT_CHECKPOINT_HOVER = wx.PyEventBinder(wxEVT_CHECKPOINT_HOVER, 1)
wxEVT_CHECKPOINT_SELECTED = wx.NewEventType()
EVT_CHECKPOINT_SELECTED = wx.PyEventBinder(wxEVT_CHECKPOINT_SELECTED, 1)
wxEVT_CHECKPOINT_RANGE_SELECTED = wx.NewEventType()
EVT_CHECKPOINT_RANGE_SELECTED = wx.PyEventBinder(wxEVT_CHECKPOINT_RANGE_SELECTED, 1)


class CheckpointTreeEvent(wx.PyCommandEvent):
    """This event class is used for all events generated by the CheckpointTree
    """
    def __init__(self, type, obj, mouse_info):
        wx.PyCommandEvent.__init__(self, type)
        self.SetEventObject(obj)
        self.SetId(obj.GetId())
        self.mouse_info = mouse_info

    def GetLine(self):
        """The RestartLine of the event
        """
        return self.mouse_info.nearest_line

    def GetRestartNumber(self):
        """The restart number of the event; only valid if there is a line
        associated with this event, otherwise will raise AttributeError
        """
        return self.mouse_info.restart_number

    def GetFrameNumber(self):
        """The frame number of the event
        """
        return self.mouse_info.frame_number


class CheckpointTreeRangeEvent(wx.PyCommandEvent):
    """This event class is used for all events generated by the CheckpointTree
    """
    def __init__(self, type, obj, start, end):
        wx.PyCommandEvent.__init__(self, type)
        self.SetEventObject(obj)
        self.SetId(obj.GetId())
        self.start = start
        self.end = end


class CheckpointTreeMouseInfo:
    def __init__(self, tree, evt):
        self.tree = tree
        self.nearest_line = None
        self.calc_mouse_info(evt)

    @property
    def is_near(self):
        return self.nearest_line is not None

    @property
    def restart_number(self):
        return self.nearest_line.restart_number

    def calc_mouse_info(self, evt):
        tree = self.tree
        ex, ey = evt.GetX(), evt.GetY()
        sx, sy = tree.GetViewStart()
        x, y = tree.CalcUnscrolledPosition(ex, ey)
        self.x, self.y = x, y
        size = tree.GetVirtualSize()
        # print(ex, sx, x)

        self.frame_number = (x - tree.width_border + tree.x_scale // 2) // tree.x_scale
        self.level = ((tree.virtual_height - y) - tree.height_border + tree.level_height // 2) // tree.level_height
        x1 = tree.frame_to_x(self.frame_number)
        y1 = tree.level_to_y(self.level)
        self.x_hit = abs(x1 - x)
        self.y_hit = abs(y1 - y)
        if self.y_hit < tree.y_hit:
            # print(frame_number, level)
            self.nearest_line = tree.restart_lines.find_line(self.frame_number, self.level)
        else:
            self.nearest_line = None


class CheckpointTree(wx.ScrolledWindow):
    def __init__(self, parent, source):
        wx.ScrolledWindow.__init__(self, parent, -1)

        self.width_border = 10
        self.height_border = 10
        self.x_scale = 1
        self.level_height = 10
        self.line_width = 3
        self.marker_size = 3
        self.x_hit = 5
        self.y_hit = 5
        self.over_line = None
        self.highlight_over_line = False
        self.scroll_chunk = 20
        self.mouse_mode = None
        self.initial_mouse_state = None
        self.current_mouse_state = None
        self.virtual_height = -1
        self.virtual_width = -1
        self.compute_lines(source)

        self.bg_color = wx.WHITE
        self.line_color = wx.BLUE
        self.marker_color = wx.Colour(255, 255, 255)
        self.cursor_color = wx.Colour(255, 0, 255)
        self.create_dc_stuff()

        self.SetBackgroundColour(self.bg_color)

        self.SetScrollRate(self.scroll_chunk, self.scroll_chunk)

        self.Bind(wx.EVT_PAINT, self.on_paint)
        self.Bind(wx.EVT_SIZE, self.on_size)
        self.Bind(wx.EVT_LEFT_DOWN, self.on_left_down)
        self.Bind(wx.EVT_MOTION, self.on_motion)
        self.Bind(wx.EVT_LEFT_UP, self.on_left_up)

    def compute_lines(self, source=None):
        if source is None:
            source = self.source
        self.source = source
        data = source.get_restart_summary()
        # print("DATA", data)
        parent_level = 0
        data.sort()
        order = get_order(data, parent_level)
        order[0:0] = [data[0]]
        # from pprint import pprint
        # pprint(order)
        lines = RestartLines(order)
        # print("restart_lines", lines)
        self.set_lines(lines)
        i = self.source.current_frame_number
        if i is not None:
            # print("current frame number:", i)
            dx = self.GetScrollPos(wx.HORIZONTAL)
            if i > dx:
                self.Scroll(i, -1)

    def recalc_view(self):
        self.compute_lines()

    def refresh_view(self):
        self.Refresh()

    def set_lines(self, lines):
        self.restart_lines = lines
        self.max_width = lines.last_frame * self.x_scale + 2 * self.width_border
        self.max_height = self.level_height * lines.highest_level + 2 * self.height_border
        if self.max_width > self.virtual_width:
            self.virtual_width = self.max_width
        if self.max_height > self.virtual_height:
            self.virtual_height = self.max_height
        self.SetVirtualSize((self.max_width, self.max_height))

    def create_dc_stuff(self):
        self.pen = wx.Pen(self.line_color, self.line_width)
        self.hover_pen = wx.Pen(self.line_color, self.line_width * 2 + 1)
        self.marker_brush = wx.Brush(self.marker_color)
        self.marker_pen = wx.Pen(self.line_color, 1)
        self.cursor_brush = wx.Brush(self.cursor_color)

    def on_size(self, event):
        size = self.GetVirtualSize()
        self.virtual_width = size.x
        self.virtual_height = size.y

    def on_paint(self, event):
        dc = wx.PaintDC(self)
        x0, y0 = self.CalcUnscrolledPosition (0, 0)
        dc.SetLogicalOrigin(x0, y0)

        dc.SetPen(self.pen)
        for line in self.restart_lines:
            x1 = self.frame_to_x(line.start_frame)
            x2 = self.frame_to_x(line.end_frame)
            y = self.level_to_y(line.level)
            if self.highlight_over_line and line == self.over_line:
                dc.SetPen(self.hover_pen)
                dc.DrawLine(x1, y, x2, y)
            else:
                dc.SetPen(self.pen)
                dc.DrawLine(x1, y, x2, y)
            dc.SetPen(self.marker_pen)
            dc.SetBrush(self.marker_brush)
            for interesting_frame in line.interesting_frames.keys():
                xi = self.frame_to_x(interesting_frame)
                dc.DrawCircle(xi, y, self.marker_size)

            if line.restart_number == 0:
                continue

            # draw vertical line connecting to parent restart
            parent_line = self.restart_lines[line.parent]
            parent_y = self.level_to_y(parent_line.level)
            dc.SetPen(self.pen)
            dc.DrawLine(x1, y, x1, parent_y)

        # highlighting of selected lines
        if self.current_mouse_state is not None:
            dc.SetPen(self.hover_pen)
            start, end = self.initial_mouse_state, self.current_mouse_state
            if start.frame_number > end.frame_number:
                start, end = end, start
            current_frame = end.frame_number
            leftmost_frame = start.frame_number
            line = end.nearest_line
            while current_frame > leftmost_frame:
                if line.is_frame_in_line(leftmost_frame):
                    left_frame = leftmost_frame
                else:
                    left_frame = line.start_frame
                x1 = self.frame_to_x(left_frame)
                x2 = self.frame_to_x(current_frame)
                y = self.level_to_y(line.level)
                dc.DrawLine(x1, y, x2, y)
                current_frame = left_frame
                if current_frame > leftmost_frame:
                    line = self.restart_lines.find_parent(line)

        # current frame & restart number
        i = self.source.current_frame_number
        if i is not None:
            xc = self.frame_to_x(i)
            restart = self.source.get_restart_of_frame(i)
            line = self.restart_lines[restart.restart_number]
            yc = self.level_to_y(line.level)
            dc.SetPen(self.marker_pen)
            dc.SetBrush(self.cursor_brush)
            dc.DrawCircle(xc, yc, self.marker_size)

    def frame_to_x(self, frame_number):
        return self.width_border + (self.x_scale * frame_number)

    def level_to_y(self, level):
        return self.virtual_height - (self.height_border + level * self.level_height)

    def on_left_down(self, evt):
        mouse = CheckpointTreeMouseInfo(self, evt)
        if mouse.is_near:
            evt = CheckpointTreeEvent(wxEVT_CHECKPOINT_SELECTED, self, mouse)
            self.GetEventHandler().ProcessEvent(evt)
            self.initial_mouse_state = mouse
            self.current_mouse_state = None
            self.mouse_mode = "range"

    def on_motion(self, evt):
        mouse = CheckpointTreeMouseInfo(self, evt)
        if mouse.is_near:
            if self.mouse_mode == "range":
                self.current_mouse_state = mouse
                evt = CheckpointTreeRangeEvent(wxEVT_CHECKPOINT_RANGE_SELECTED, self, self.initial_mouse_state, mouse)
                self.GetEventHandler().ProcessEvent(evt)
            else:
                evt = CheckpointTreeEvent(wxEVT_CHECKPOINT_HOVER, self, mouse)
                self.GetEventHandler().ProcessEvent(evt)
                if self.over_line != mouse.nearest_line:
                    # print("REFRESHING")
                    self.over_line = mouse.nearest_line
                    wx.CallAfter(self.Refresh)

    def on_left_up(self, evt):
        self.mouse_mode = None
        evt.Skip()


#For testing 
if __name__ == '__main__':
    app = wx.App()
    frame = wx.Frame(None, -1, "Test CheckpointTree", size=(500,400))

    from pprint import pprint

    data.sort()
    pprint(data)

    min_frame = min([d[1] for d in data])
    pprint(min_frame)

    max_frame = max([d[3] for d in data])
    pprint(max_frame)

    parent_level = 0
    order = get_order(data, parent_level)
    order[0:0] = [data[0]]
    pprint(order)

    lines = RestartLines(order)
    print("restart_lines", lines)

    class TestSource:
        current_frame_number = None
        def get_restart_summary(self):
            return list(data)

    scroll = CheckpointTree(frame, TestSource())
    frame.Show() 
    app.MainLoop()



import os
import sys

import wx
import numpy as np

from sawx.utils.nputil import intscale
from sawx.ui import compactgrid as cg

from atrip.machines import atari8bit

from ..viewer import SegmentViewer

import logging
log = logging.getLogger(__name__)


class AnticColorViewer(SegmentViewer):
    has_colors = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._antic_color_registers = list(atari8bit.powerup_colors())
        self._color_standard_name = "NTSC"
        self._color_standard = None
        self._color_registers = None

    @property
    def antic_color_registers(self):
        return self._antic_color_registers

    @antic_color_registers.setter
    def antic_color_registers(self, value):
        baseline = list(atari8bit.powerup_colors())
        # need to operate on a copy of the colors to make sure we're not
        # changing some global value. Also force as python int so we're not
        # mixing numpy and python values.
        if len(value) == 5:
            baseline[4:9] = [int(i) for i in value]
        else:
            baseline[0:len(value)] = [int(i) for i in value]
        self._antic_color_registers = baseline
        self._color_registers = None
        print("ANTIC COLOR REGISTERS", self._antic_color_registers)
        self.colors_changed()
        self.graphics_properties_changed()

    @property
    def color_standard_name(self):
        return self._color_standard_name

    @color_standard_name.setter
    def color_standard_name(self, value):
        self._color_standard_name = value
        self._color_standard = None
        self._color_registers = None
        self.colors_changed()
        self.graphics_properties_changed()

    @property
    def color_standard(self):
        if self._color_standard is None:
            self._color_standard = atari8bit.valid_color_standards[self.color_standard_name]
        return self._color_standard

    @property
    def color_registers(self):
        if self._color_registers is None:
            self._color_registers = atari8bit.get_color_registers(self.antic_color_registers, self.color_standard)
        return self._color_registers

    def colors_changed(self):
        """Hook for subclasses that need to invalidate stuff when colors change
        """
        pass

    def graphics_properties_changed(self):
        print(f"graphics_properties_changed! std={self.color_standard_name}")
        self.control.recalc_view()
        self.control.refresh_view()



import lzma
import io

import numpy as np

from .. import errors
from ..compressor import Compressor


class LZMACompressor(Compressor):
    compression_algorithm = "lzma"

    def calc_unpacked_data(self, byte_data):
        try:
            buf = io.BytesIO(byte_data)
            with lzma.LZMAFile(buf, mode='rb') as f:
                unpacked = f.read()
        except lzma.LZMAError as e:
            raise errors.InvalidAlgorithm(e)
        if len(unpacked) == 0:
            raise errors.InvalidAlgorithm("Unpacked to zero size")
        return unpacked

    def calc_packed_data(self, byte_data, media=None):
        buf = io.BytesIO()
        try:
            with lzma.LZMAFile(buf, mode='wb') as f:
                f.write(byte_data)
        except OSError as e:
            raise errors.InvalidAlgorithm(e)
        else:
            packed = buf.getvalue()
        return packed



# generic6502 input template locations
import numpy as np

VIDEO_WIDTH = 40
VIDEO_HEIGHT = 192

VIDEO_SIZE = VIDEO_WIDTH * VIDEO_HEIGHT
AUDIO_SIZE = 2048
MAIN_MEMORY_SIZE = 1<<16
STATESAV_MAX_SIZE = 256 + MAIN_MEMORY_SIZE + 256


OUTPUT_DTYPE = np.dtype([
    ("state", np.uint8, STATESAV_MAX_SIZE),
    ("video", np.uint8, VIDEO_SIZE),
    ("scan_line_flags", np.uint8, VIDEO_HEIGHT),
    ("audio", np.uint8, AUDIO_SIZE),
])

STATESAV_DTYPE = np.dtype([
    ("PC", '<u2'),
    ("A", np.uint8),
    ("X", np.uint8),
    ("Y", np.uint8),
    ("SP", np.uint8),
    ("P", np.uint8),
    ("unused1", np.uint8, 249),
    ("memory", np.uint8, MAIN_MEMORY_SIZE),
    ("hires_graphics", np.uint8),
    ("text_mode", np.uint8),
    ("mixed_mode", np.uint8),
    ("alt_page_select", np.uint8),
    ("tv_line", np.uint8),
    ("tv_cycle", np.uint8),
    ("unused2", np.uint8, 250),
    ])

CPU_DTYPE = np.dtype([
    ("PC", '<u2'),
    ("A", np.uint8),
    ("X", np.uint8),
    ("Y", np.uint8),
    ("SP", np.uint8),
    ("P", np.uint8),
    ("unused1", np.uint8, 249),
    ])



import numpy as np

from .. import errors
from ..media_type import DiskImage
from ..segment import Segment
from ..container import ContainerHeader

import logging
log = logging.getLogger(__name__)


class CassetteChunk:
    def __init__(self, media, index):
        header = media[index:index + 8]
        if len(header) < 8:
            raise errors.InvalidSectorNumber(f"tape position {index} beyond end of tape")
        self.tape_index = index
        flag = header[0:4].view(dtype='>u4')
        self.chunk_length = int(header[4] + header[5] * 256)
        self.chunk_aux = int(header[6] + header[7] * 256)
        self.record_length = self.chunk_length + 8
        if flag == 0x46554a49:  # FUJI in big-endian hex
            self.chunk_type = "FUJI"
        elif flag == 0x62617564:  # baud in big-endian hex
            self.chunk_type = "baud"
        elif flag == 0x64617461:  # daha in big-endian hex
            self.chunk_type = "data"
        else:
            try:
                name = flag.decode("latin1")
            except:
                name = hex(flag)
            raise errors.UnsupportedSectorType(f"Can't handle {name} chunk types")
        self.chunk_data = media[index + 8:index + 8 + self.chunk_length]
        if len(self.chunk_data) < self.chunk_length:
            raise errors.InvalidSectorNumber(f"chunk {self.chunk_type} at tape index {index} specifies {self.chunk_length} bytes but ran out of tape")

    def __str__(self):
        return f"{self.chunk_type}: {self.chunk_length} bytes at tape index {self.tape_index}"


class AtariCassetteImage(DiskImage):
    ui_name = "Atari Cassette Image (.cas)"

    def check_media_size(self):
        size = len(self)
        if size < 8:
            raise errors.InvalidMediaSize(f"{self.ui_name} too small to be valid image")

    def check_magic(self):
        flag = self[0:4].view(dtype='>u4')
        if flag != 0x46554a49:  # FUJI in big-endian hex
            raise errors.InvalidHeader("Not a .cas image; FUJI chunk not found")

    def get_chunk(self, index):
        return CassetteChunk(self, index)



import sys

import numpy as np
import wx
import wx.lib.scrolledpanel as scrolled
import wx.lib.buttons as buttons

from sawx.ui.tilelist import TileButton

import logging
log = logging.getLogger(__name__)


class TileWrapControl(wx.Panel):
    """
    View for displaying categories of tiles to paint the map
    """

    def __init__(self, parent, linked_base, command=None, **kwargs):
        wx.Panel.__init__(self, parent, -1, **kwargs)
        self.linked_base = linked_base
        self.segment_viewer = None  # Filled in at time of segment viewer creation
        self.command_cls = command

        self.cat = wx.Choice(self, -1)
        self.cat.Bind(wx.EVT_CHOICE, self.on_category)
        self.panel = scrolled.ScrolledPanel(self, -1, style=wx.VSCROLL)
        self.panel.Bind(wx.EVT_SIZE, self.panel_size)
        psiz = wx.BoxSizer(wx.VERTICAL)
        self.panel.SetSizer(psiz)
        self.panel.ShowScrollbars(wx.SHOW_SB_NEVER, wx.SHOW_SB_ALWAYS)
        self.panel.SetupScrolling(scroll_x=False)
        self.bg = wx.SystemSettings.GetColour(wx.SYS_COLOUR_BTNFACE)
        self.panel.SetBackgroundColour(self.bg)
        self.SetBackgroundColour(self.bg)

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self.cat, 0, wx.EXPAND, 0)
        sizer.Add(self.panel, 1, wx.EXPAND, 0)
        self.SetSizer(sizer)
        self.Fit()

        self.pattern_to_item = {}
        self.parse_tile_map(self.panel, [])
        self.current_tile = None
        self.setup_tiles()
        self.zoom = 2

    @property
    def editor(self):
        return self.segment_viewer.linked_base.editor

    @property
    def task(self):
        return self.segment_viewer.linked_base.editor.task

    def panel_size(self, evt):
        size = self.panel.GetSize()
        vsize = self.panel.GetVirtualSize()
        self.panel.SetVirtualSize((size[0] - 20, vsize[1]))

    def recalc_view(self):
        if self.segment_viewer is not None:
            self.parse_tile_map(self.panel, self.segment_viewer.tile_groups)
            self.setup_tiles()
            TileButton.set_colors(self.segment_viewer.machine)

    def parse_tile_map(self, panel, tile_map):
        sizer = panel.GetSizer()
        sizer.Clear(True)
        self.tile_map = tile_map
        self.categories = []
        self.items = []
        self.pattern_to_item = {}
        for items in tile_map:
            label = items[0]
            t = wx.StaticText(panel, -1, label)
            sizer.Add(t, 0, wx.EXPAND, 0)
            self.categories.append(t)
            w = wx.WrapSizer()
            for tiles in items[1:]:
                for i in np.arange(np.alen(tiles)):
                    data = tiles[i:i+1]
                    bmp = self.segment_viewer.machine.antic_font.get_image(data[0], self.zoom)
                    btn = TileButton(panel, -1, bmp, style=wx.BORDER_NONE|wx.BU_EXACTFIT)
                    btn.SetBackgroundColour(self.bg)
                    btn.tile_data = data
                    btn.Bind(wx.EVT_BUTTON, self.on_tile_clicked)
                    w.Add(btn, 0, wx.ALL, 0)
                    self.items.append(btn)
                    self.pattern_to_item[tuple(data)] = btn
            sizer.Add(w, 0, wx.EXPAND, 0)
        self.Layout()

    def setup_tiles(self):
        cats = [str(c.GetLabelText()) for c in self.categories]
        self.cat.Set(cats)
        self.cat.SetSelection(0)

    def scroll_to_control(self, ctrl):
        sppu_x, sppu_y = self.panel.GetScrollPixelsPerUnit()
        vs_x, vs_y = self.panel.GetViewStart()
        cr = ctrl.GetRect()
        self.panel.Scroll(0, vs_y + (cr.y / sppu_y))

    def on_category(self, event):
        cat = event.GetSelection()
        label = self.categories[cat]
        self.scroll_to_control(label)

    def on_tile_clicked(self, event):
        btn = event.GetEventObject()
        self.current_tile = btn
        self.clear_toggle_except(btn)
        wx.CallAfter(self.segment_viewer.set_draw_pattern, btn.tile_data)

    def show_pattern(self, pattern):
        log.debug("tilelist showing pattern %s" % str(pattern))
        btn = self.pattern_to_item.get((int(pattern),), None)
        self.clear_toggle_except(btn)
        if btn is not None:
            self.scroll_to_control(btn)

    def clear_tile_selection(self):
        self.current_tile = None
        self.clear_toggle_except()

    def clear_toggle_except(self, btn=None):
        for b in self.items:
            if b != btn:
                b.SetToggle(False)
        if btn is not None:
            btn.SetToggle(True)

    def on_tile(self, event):
        index = event.GetInt()
        self.change_tile(self.items[index])

    def change_tile(self, tile):
        e = self.editor
        if e is None:
            return
        cmd_cls = self.command_cls
        if cmd_cls is None:
            return
        if e.can_copy:
            index = e.anchor_start_index
        else:
            index = e.caret_index
        value = tile.get_bytes()
        cmd = cmd_cls(e.segment, index, index+len(value), value, True)
        e.process_command(cmd)


if __name__ == '__main__':
    import sys
    sys.path[0:0] = [".."]
    print(sys.path)
    import omnivore.arch.fonts as fonts
    from atrip.machines import atari8bit 

    class Wrapper(object):
        def __init__(self, **kwargs):
            for k, v in kwargs.items():
                setattr(self, k, v)

        def set_current_draw_pattern(self, *args, **kwargs):
            pass

    class MyFrame(wx.Frame):
        def __init__(self, parent, id, title):
            wx.Frame.__init__(self, parent, id, title, wx.DefaultPosition, wx.DefaultSize)
            tile_map = [
            ("road", [0x70]),
            ("trees", list(range(0x80, 0x96)), list(range(0x01, 0x16)),),
            ("buildings", list(range(0x96, 0x9f)), list(range(0x16, 0x1f)), list(range(0x41, 0x51)), list(range(0x5d, 0x60)),),
            ("people", list(range(0xf1, 0xf4)), list(range(0x71, 0x74))),
            ("water", list(range(0x2e, 0x41)),),
            ("bridges", list(range(0x69, 0x6d)),),
            ("vehicles", list(range(0x51, 0x59)),),
            ("airport", list(range(0x60, 0x68)), [0x5f], list(range(0x59, 0x5d)), list(range(0xd9, 0xdd))),
            ("golf", list(range(0xa9, 0xae)),),
            ("other", [0x20, 0x25, 0x26, ]),
            ("special", list(range(0x21, 0x25)), list(range(0x74, 0x76)),),
                ]
            color_converter = atari8bit.gtia_ntsc_to_rgb
            highlight_color = (100, 200, 230)
            unfocused_caret_color = (128, 128, 128)
            background_color = (255, 255, 255)
            match_background_color = (255, 255, 180)
            comment_background_color = (255, 180, 200)
            antic_font = fonts.AnticFont(fonts.A8DefaultFont, 4, atari8bit.powerup_colors(), highlight_color, match_background_color, comment_background_color, color_converter)
            editor = Wrapper(antic_font=antic_font, antic_tile_map=tile_map, highlight_color=highlight_color)
            task = Wrapper(active_editor=editor)
            panel = TileWrapControl(self, task)
            panel.recalc_view()

    class MyApp(wx.App):
        def OnInit(self):
            frame = MyFrame(None, -1, 'test')
            frame.Show(True)
            frame.Center()
            return True

    app = MyApp(0)
    app.MainLoop()



import glob

import numpy as np

from mock import *

from atrip.container import guess_container
from atrip.media_type import Media, guess_media_type
from atrip import errors

from atrip.media_types.atari_disks import *
from atrip.media_types.apple_disks import *


class TestAtariDos2:
    base_path = None
    expected_mime = ""

    @pytest.mark.parametrize("pathname", sorted(glob.glob(os.path.join(os.path.dirname(__file__), "../samples/", "*"))))
    def test_samples_dir(self, pathname):
        # wrapped, ext = os.path.splitext(pathname)
        # print(f"checking {pathname}")
        # sample_data = np.fromfile(pathname, dtype=np.uint8)
        # container = guess_container(sample_data)
        # if container.compression_algorithm != "no compression":
        #     _, ext = os.path.splitext(wrapped)
        # container.guess_media_type()
        # print(ext, ext_to_valid_types)
        # if ext in ext_to_valid_types:
        #     assert container.media.__class__ in ext_to_valid_types[ext]
        # else:
        #     assert container.media.__class__ == Media
        if ".tar" in pathname or ".zip" in pathname:
            pytest.skip(f"skipping archive collections for this test: {pathname}")
        sample_data = np.fromfile(pathname, dtype=np.uint8)
        if len(sample_data) == 0:
            pytest.skip(f"skipping zero-length files for this test: {pathname}")
        container = guess_container(sample_data)
        container.guess_media_type()
        is_expected_media(container, pathname)


if __name__ == "__main__":
    import logging
    logging.basicConfig(level=logging.WARNING)
    log = logging.getLogger("atrip.media_type")
    log.setLevel(logging.DEBUG)

    def check(pathname):
        print(f"checking {pathname}")
        sample_data = np.fromfile(pathname, dtype=np.uint8)
        container = guess_container(sample_data)
        container.guess_media_type()
        container.guess_filesystem()
        print(container.media.filesystem)
        print(container.verbose_info)

    import sys
    import glob
    if len(sys.argv) > 1:
        images = sys.argv[1:]
    else:
        images = sorted(glob.glob(os.path.join(os.path.dirname(__file__), "../samples/", "*")))
    for pathname in images:
        check(pathname)



""" Pixel converters

ByteValidator: change byte index list into rectangular grid
ColorIndexGenerator: bytes to pixel color indexes

calc_rgb_from_color_indexes: create RGB array given pixels, styles, and RGB colro mappings


Utility classes/functions to convert byte representations of pixel data into
grids of color register indexes.

Color register indexes are byte values, each byte calculated from the integer
value of the bits making up the corresponding pixel. For example, Antic mode D
(graphics 7) uses 2 bits per pixel, laid out into 4 pixels:

  bit7 bit6 | bit5 bit4 | bit3 bit2 | bit1 bit0

where the color register indexes for each of the pixels is computed by:

  bit7*2 + bit6
  bit5*2 + bit4
  bit3*2 + bit2
  bit1*2 + bit0

Each byte would therefore contain values in the range of [0, 3].

This conversion is device independent; the code that displays pixels on screen will take these values and convert them to the color values referred to by these indexes.
"""
import numpy as np

from atrip import style_bits
ignore_mask = style_bits.not_user_bit_mask & (0xff ^ style_bits.diff_bit_mask)
invalid_style = 0xff

from sawx.utils.permute import bit_reverse_table


class ConverterBase:
    name = "base"
    ui_name = "Base"

    pixels_per_byte = 8
    bitplanes = 1

    scale_width = 1
    scale_height = 1

    @classmethod
    def validate_pixels_per_row(cls, pixels_per_row):
        return (pixels_per_row // cls.pixels_per_byte) * cls.pixels_per_byte

    @classmethod
    def calc_bytes_per_row(cls, pixels_per_row):
        return (pixels_per_row + cls.pixels_per_byte - 1) // cls.pixels_per_byte

    @classmethod
    def calc_grid_height(cls, num_byte_values, bytes_per_row):
        return (num_byte_values + bytes_per_row - 1) // bytes_per_row

    def calc_color_index_grid(self, byte_values, style, bytes_per_row):
        nr = len(byte_values) // bytes_per_row
        pixels = self.calc_pixels(byte_values, bytes_per_row)
        style_per_pixel = self.calc_style_per_pixel(pixels, style)
        return pixels.reshape((-1, bytes_per_row * self.pixels_per_byte)), style_per_pixel.reshape((-1, bytes_per_row * self.pixels_per_byte))


class Converter1bpp(ConverterBase):
    name = "1bpp"
    ui_name = "1 bit per pixel"
    
    pixels_per_byte = 8

    def calc_pixels(self, byte_values, bytes_per_row):
        bits = np.unpackbits(byte_values)
        pixels = bits.reshape((-1, 8))
        return pixels

    def calc_style_per_pixel(self, pixels, style):
        h, w = pixels.shape
        stack = np.empty((len(style), 8), dtype=style.dtype)
        stack[:,0] = style
        stack[:,1] = style
        stack[:,2] = style
        stack[:,3] = style
        stack[:,4] = style
        stack[:,5] = style
        stack[:,6] = style
        stack[:,7] = style
        return stack.reshape((h, w))


class Converter2bpp(ConverterBase):
    name = "2bpp"
    ui_name = "2 bits per pixel"
    
    pixels_per_byte = 4

    def calc_pixels(self, byte_values, bytes_per_row):
        bits = np.unpackbits(byte_values)
        bits = bits.reshape((-1, 8))
        pixels = np.empty((bits.shape[0], 4), dtype=np.uint8)
        pixels[:,0] = bits[:,0] * 2 + bits[:,1]
        pixels[:,1] = bits[:,2] * 2 + bits[:,3]
        pixels[:,2] = bits[:,4] * 2 + bits[:,5]
        pixels[:,3] = bits[:,6] * 2 + bits[:,7]
        return pixels

    def calc_style_per_pixel(self, pixels, style):
        h, w = pixels.shape
        stack = np.empty((len(style), 4), dtype=style.dtype)
        stack[:,0] = style
        stack[:,1] = style
        stack[:,2] = style
        stack[:,3] = style
        return stack.reshape((h, w))


class Converter4bpp(ConverterBase):
    name = "4bpp"
    ui_name = "4 bits per pixel"
    
    pixels_per_byte = 2

    def calc_pixels(self, byte_values, bytes_per_row):
        bits = np.unpackbits(byte_values)
        bits = bits.reshape((-1, 8))
        pixels = np.empty((bits.shape[0], 2), dtype=np.uint8)
        pixels[:,0] = bits[:,0] * 8 + bits[:,1] * 4 + bits[:,2] * 2 + bits[:,3]
        pixels[:,1] = bits[:,4] * 8 + bits[:,5] * 4 + bits[:,6] * 2 + bits[:,7]
        return pixels

    def calc_style_per_pixel(self, pixels, style):
        h, w = pixels.shape
        stack = np.empty((len(style), 2), dtype=style.dtype)
        stack[:,0] = style
        stack[:,1] = style
        return stack.reshape((h, w))


class Converter8bpp(ConverterBase):
    name = "8bpp"
    ui_name = "8 bits per pixel"
    
    pixels_per_byte = 1

    def calc_pixels(self, byte_values, bytes_per_row):
        return byte_values.reshape((-1, 1))

    def calc_style_per_pixel(self, pixels, style):
        h, w = pixels.shape
        style.reshape((h, w))
        return style

    def calc_valid_byte_grid(self, byte_values, style, grid_start_index, data_start_index, num_bytes, byte_width):
        num_prepend = max(data_start_index - grid_start_index, 0)
        _, num_extra = divmod(num_prepend + num_bytes, byte_width)
        if num_extra > 0:
            num_append = byte_width - num_extra
        else:
            num_append = 0
        if num_prepend + num_append > 0:
            total = num_prepend + num_append + num_bytes
            b = np.empty(total, dtype=byte_values.dtype)
            b[0:num_prepend] = 0
            b[num_prepend:num_prepend + num_bytes] = byte_values
            b[num_prepend + num_bytes:-1] = 0
            s[0:num_prepend] = 0
            s[num_prepend:num_prepend + num_bytes] = style
            s[num_prepend + num_bytes:-1] = 0
        else:
            b = byte_values
            s = style
        return b.reshape((-1, byte_width), s.reshape((-1, byte_width)))


class AnticB(Converter1bpp):
    name = "antic_b"
    ui_name = "ANTIC Mode B (Gr 6, 1bpp)"

    scale_width = 2
    scale_height = 2


class AnticC(AnticB):
    name = "antic_c"
    ui_name = "ANTIC Mode C (Gr 6+, 1bpp)"

    scale_width = 2
    scale_height = 1


class AnticD(Converter2bpp):
    name = "antic_d"
    ui_name = "ANTIC Mode D (Gr 7, 2bpp)"

    scale_width = 2
    scale_height = 2


class AnticE(AnticD):
    name = "antic_e"
    ui_name = "ANTIC Mode E (Gr 7+, 2bpp)"

    scale_width = 2
    scale_height = 1


class AnticNormalPlayer(AnticC):
    name = "Player/Missile, normal width"
    scale_width = 2


class AnticDoublePlayer(AnticNormalPlayer):
    name = "Player/Missile, double width"
    scale_width = 4


class AnticQuadPlayer(AnticNormalPlayer):
    name = "Player/Missile, quad width"
    scale_width = 8


def calc_rgb_from_color_indexes_python(color_indexes, style_per_pixel, colors, empty_color):
    h, w = color_indexes.shape
    color_indexes = color_indexes.reshape(-1)
    style_per_pixel = style_per_pixel.reshape(-1)
    flat_image = np.empty((h * w, 3), dtype=np.uint8)
    color_registers, h_colors, m_colors, c_colors, d_colors = colors
    for i in range(len(color_indexes)):
        color_index = color_indexes[i]
        style = style_per_pixel[i]
        if style == invalid_style:
            flat_image[i] = empty_color
        elif style & ignore_mask == 0:
            flat_image[i] = color_registers[color_index]
        elif style & style_bits.selected_bit_mask:
            flat_image[i] = h_colors[color_index]
        elif (style & style_bits.user_bit_mask) > 0:
            flat_image[i] = d_colors[color_index]
        elif style & style_bits.comment_bit_mask:
            flat_image[i] = h_colors[color_index]
        elif style & style_bits.match_bit_mask:
            flat_image[i] = h_colors[color_index]
        else:
            flat_image[i] = (0xff, 0, 0xee)  # not any of the above?
    return flat_image.reshape((h, w, 3))

try:
    from . import pixel_speedups
    calc_rgb_from_color_indexes = pixel_speedups.calc_rgb_from_color_indexes
except ImportError:
    calc_rgb_from_color_indexes = calc_rgb_from_color_indexes_python



import zipfile

import numpy as np

from . import errors
from .segments import SegmentData, EmptySegment, ObjSegment
from .diskimages import DiskImageBase
from .utils import to_numpy

import logging
log = logging.getLogger(__name__)


class MameZipImage(DiskImageBase):
    def __init__(self, rawdata, filename=""):
        self.zipdata = rawdata
        fh = self.zipdata.bufferedio
        if zipfile.is_zipfile(fh):
            with zipfile.ZipFile(fh) as zf:
                self.check_zip_size(zf)
                self.create_rawdata(zf)
        else:
            raise errors.InvalidDiskImage("Not a MAME zip file")
        DiskImageBase.__init__(self, self.rawdata, filename)

    def __str__(self):
        return "MAME Zip file, %d ROMs, orig_size=%d, uncompressed=%d" % (len(self.zip_segment_info), len(self.zipdata), len(self.rawdata))

    def setup(self):
        self.check_size()

    def strict_check(self):
        pass

    def relaxed_check(self):
        pass

    def check_zip_size(self, zf):
        for item in zf.infolist():
            _, r = divmod(item.file_size, 16)
            if r > 0:
                raise errors.InvalidDiskImage("zip entry not 16 byte multiple")

    def create_rawdata(self, zf):
        roms = []
        segment_info = []
        offset = 0
        for item in zf.infolist():
            rom = np.frombuffer(zf.open(item).read(), dtype=np.uint8)
            roms.append(rom)
            segment_info.append((offset, item.file_size, item.filename, item.CRC))
            offset += item.file_size
        data = np.concatenate(roms)
        self.zip_segment_info = segment_info
        self.rawdata = SegmentData(data)

    def check_size(self):
        pass

    def parse_segments(self):
        r = self.rawdata
        self.segments = []
        for offset, size, name, crc in self.zip_segment_info:
            end = offset + size
            self.segments.append(ObjSegment(r[offset:end], 0, offset, offset, end, name=name))
